import pytest

from ape.exceptions import SignatureError
from ape.pytest.contextmanagers import RevertsContextManager as reverts


def test_default_sender_test_account(solidity_contract_instance, owner, accounts):
    with accounts.use_sender(owner):
        tx = solidity_contract_instance.setNumber(1)
        assert tx.transaction.sender == owner.address
    with pytest.raises(SignatureError):
        solidity_contract_instance.setNumber(2)

    with accounts.use_sender(owner.address):
        tx = solidity_contract_instance.setNumber(1)
        assert tx.transaction.sender == owner.address

    with accounts.use_sender(owner.index):
        tx = solidity_contract_instance.setNumber(1)
        assert tx.transaction.sender == owner.address


def test_default_sender_account(
    project,
    networks_connected_to_tester,
    account_manager,
    keyfile_account,
):
    passphrase = "asdf1234"

    with account_manager.use_sender(keyfile_account) as acct:
        acct.set_autosign(True, passphrase)
        contract = keyfile_account.deploy(project.SolidityContract, 0)

    with account_manager.use_sender(keyfile_account) as acct:
        acct.set_autosign(True, passphrase)
        tx = contract.setNumber(1)
        assert tx.transaction.sender == keyfile_account.address

    with pytest.raises(SignatureError):
        contract.setNumber(2)

    with account_manager.use_sender(keyfile_account.address) as acct:
        acct.set_autosign(True, passphrase)
        tx = contract.setNumber(1)
        assert tx.transaction.sender == keyfile_account.address

    with account_manager.use_sender(keyfile_account.alias) as acct:
        acct.set_autosign(True, passphrase)
        tx = contract.setNumber(1)
        assert tx.transaction.sender == keyfile_account.address

    with account_manager.use_sender(0) as acct:
        acct.set_autosign(True, passphrase)
        tx = contract.setNumber(1)
        assert tx.transaction.sender == keyfile_account.address


def test_nested_default_sender(solidity_contract_instance, owner, accounts, not_owner):
    with accounts.use_sender(owner):
        tx = solidity_contract_instance.setNumber(1)
        assert tx.transaction.sender == owner.address
        with accounts.use_sender(not_owner):
            with reverts():
                solidity_contract_instance.setNumber(2)

        solidity_contract_instance.setNumber(3)
        assert tx.transaction.sender == owner.address


def test_with_error(solidity_contract_instance, account_manager, not_owner):
    # safe to use reverts with use_sender and when outside of the use_sender
    # there is no remaining default_user set
    with reverts("!authorized"):
        with account_manager.test_accounts.use_sender(not_owner):
            solidity_contract_instance.setNumber(2)

    assert account_manager.default_sender is None



import sys
from collections.abc import Iterable
from pathlib import Path
from typing import Any

import click
import pytest
from click import Command

from ape.cli.options import ape_cli_context
from ape.logging import LogLevel, _get_level


def _validate_pytest_args(*pytest_args) -> list[str]:
    threshold = len(pytest_args) - 1
    args_iter = iter(pytest_args)
    valid_args = []
    for idx, argument in enumerate(args_iter):
        if idx >= threshold:
            # If the last arg is -v without a value, it is a valid
            # pytest arg.
            valid_args.append(argument)
            break

        elif argument == "-v":
            # Ensure this is a pytest -v and not ape's -v.
            next_arg = next(args_iter)
            lvl_name = _get_level(next_arg)
            if not _is_ape_loglevel(lvl_name):
                valid_args.append(argument)

        else:
            valid_args.append(argument)

    return valid_args


def _is_ape_loglevel(value: Any) -> bool:
    if isinstance(value, int | LogLevel):
        return True

    elif isinstance(value, str):
        return (
            value.upper() in [x.name for x in LogLevel]
            or (value.isnumeric() and int(value) in LogLevel)
            or value.lower().startswith("loglevel.")
        )

    return False


class ApeTestCommand(Command):
    def parse_args(self, ctx, args: list[str]) -> list[str]:
        num_args = len(args)
        for idx, argument in enumerate(args):
            if not argument.startswith("-v"):
                continue
            elif (idx == num_args - 1) or argument in ("-vv", "-vvv"):
                # Definitely for pytest.
                ctx.obj["pytest_verbosity"] = argument
                args = [a for a in args if a != argument]
            else:
                # -v with a following arg; ensure not Ape's.
                next_arg = args[idx + 1]
                if not _is_ape_loglevel(next_arg):
                    ctx.obj["pytest_verbosity"] = "-v"
                    args = [a for a in args if a != argument]

        return super().parse_args(ctx, args)


@click.command(
    add_help_option=False,  # NOTE: This allows pass-through to pytest's help
    short_help="Launches pytest and runs the tests for a project",
    context_settings=dict(ignore_unknown_options=True),
    cls=ApeTestCommand,
)
# NOTE: Using '.value' because more performant.
@ape_cli_context(
    default_log_level=LogLevel.WARNING.value,
)
@click.option(
    "-w",
    "--watch",
    is_flag=True,
    default=False,
    help="Watch for changes to project files and re-run the test suite with the given options.",
)
@click.option(
    "--watch-folders",
    multiple=True,
    type=Path,
    default=[Path("contracts"), Path("tests")],
    help=(
        "Folders to watch for changes using `ape test --watch`."
        " Defaults to `contracts/` and `tests/`"
    ),
)
@click.option(
    "--watch-delay",
    type=float,
    default=0.5,
    help="Delay between polling cycles for `ape test --watch`. Defaults to 0.5 seconds.",
)
@click.argument("pytest_args", nargs=-1, type=click.UNPROCESSED)
def cli(cli_ctx, watch, watch_folders, watch_delay, pytest_args):
    pytest_arg_ls = [*pytest_args]
    if pytest_verbosity := cli_ctx.get("pytest_verbosity"):
        pytest_arg_ls.append(pytest_verbosity)

    pytest_arg_ls = _validate_pytest_args(*pytest_arg_ls)
    if watch:
        _run_with_observer(watch_folders, watch_delay, *pytest_arg_ls)

    else:
        return_code = pytest.main([*pytest_arg_ls], ["ape_test"])
        if return_code:
            # only exit with non-zero status to make testing easier
            sys.exit(return_code)


def _run_with_observer(watch_folders: Iterable[Path], watch_delay: float, *pytest_arg_ls: str):
    # Abstracted for testing purposes.
    from ape_test._watch import run_with_observer as run

    run(watch_folders, watch_delay, *pytest_arg_ls)



from typing import TYPE_CHECKING, Optional

import pytest

from ape.exceptions import BlockNotFoundError
from ape.pytest.fixtures import IsolationManager, PytestApeFixtures
from ape.pytest.utils import Scope

if TYPE_CHECKING:
    from ape.types.vm import SnapshotID


@pytest.fixture
def config_wrapper(mocker):
    return mocker.MagicMock()


@pytest.fixture
def receipt_capture(mocker):
    return mocker.MagicMock()


@pytest.fixture
def isolation_manager(config_wrapper, receipt_capture):
    return IsolationManager(config_wrapper, receipt_capture)


@pytest.fixture
def fixtures(mocker, isolation_manager):
    return PytestApeFixtures(mocker.MagicMock(), isolation_manager)


@pytest.fixture
def isolation(fixtures):
    return fixtures.isolation_manager.isolation(Scope.FUNCTION)


@pytest.fixture
def mock_evm(mocker):
    return mocker.MagicMock()


def test_isolation(isolation, receipt_capture):
    # Set up receipt capture to fail on __exit__
    # AFTER the yield statement. There was a bug
    # where we got a double-yield in this situation.

    receipt_capture.__exit__.side_effect = BlockNotFoundError(0)

    assert next(isolation) is None
    with pytest.raises(StopIteration):
        next(isolation)


def test_isolation_restore_not_implemented(mocker, networks, fixtures):
    isolation = fixtures.isolation_manager.isolation(Scope.FUNCTION)
    mock_provider = mocker.MagicMock()
    mock_provider.restore.side_effect = NotImplementedError
    mock_provider.snapshot.return_value = 123
    orig_provider = networks.active_provider
    networks.active_provider = mock_provider
    fixtures.isolation_manager.supported = True

    try:
        _ = next(isolation)
        assert mock_provider.snapshot.call_count == 1
        with pytest.raises(StopIteration):
            _ = next(isolation)

        # Is false because of the not-implemented error side-effect.
        assert fixtures.isolation_manager.supported is False

        isolation = fixtures.isolation_manager.isolation(Scope.FUNCTION)
        _ = next(isolation)
        # It does not call snapshot again.
        assert mock_provider.snapshot.call_count == 1
        with pytest.raises(StopIteration):
            _ = next(isolation)

        assert mock_provider.snapshot.call_count == 1

    finally:
        networks.active_provider = orig_provider


@pytest.mark.parametrize("snapshot_id", (0, 1, "123", (b"123", 1)))
def test_isolation_snapshot_id_types(snapshot_id, fixtures):
    class IsolationManagerWithCustomSnapshot(IsolationManager):
        take_call_count = 0
        restore_call_count = 0
        restore_called_with = []

        def take_snapshot(self) -> Optional["SnapshotID"]:
            self.take_call_count += 1
            return snapshot_id

        def restore(self, scope: Scope):
            self.restore_call_count += 1
            self.restore_called_with.append(scope)

    isolation_manager = IsolationManagerWithCustomSnapshot(
        fixtures.isolation_manager.config_wrapper,
        fixtures.isolation_manager.receipt_capture,
    )
    isolation_context = isolation_manager.isolation(Scope.FUNCTION)
    next(isolation_context)  # Enter.
    assert isolation_manager.take_call_count == 1
    assert isolation_manager.restore_call_count == 0
    next(isolation_context, None)  # Exit.
    assert isolation_manager.restore_called_with == [Scope.FUNCTION]


def test_isolation_when_snapshot_fails_avoids_restore(mocker, fixtures):
    class IsolationManagerFailingAtSnapshotting(IsolationManager):
        take_called = False

        def take_snapshot(self) -> Optional["SnapshotID"]:
            self.take_called = True
            raise NotImplementedError()

    restore_spy = mocker.spy(IsolationManagerFailingAtSnapshotting, "restore_snapshot")
    isolation_manager = IsolationManagerFailingAtSnapshotting(
        fixtures.isolation_manager.config_wrapper,
        fixtures.isolation_manager.receipt_capture,
    )
    isolation_context = isolation_manager.isolation(Scope.FUNCTION)
    next(isolation_context)  # Enter.
    assert isolation_manager.take_called
    assert not restore_spy.call_count
    next(isolation_context, None)  # Exit.
    # It doesn't even try!
    assert not restore_spy.call_count


def test_isolation_restore_fails_avoids_snapshot_next_time(fixtures):
    chain_snapshots = {}

    class IsolationManagerFailingAtRestoring(IsolationManager):
        take_called = False
        restore_called = False

        def take_snapshot(self) -> Optional["SnapshotID"]:
            self.take_called = True
            chain_snapshots[self.provider.chain_id] = ["123"]
            return "123"

        def _restore(self, snapshot_id: "SnapshotID"):
            self.restore_called = True
            raise NotImplementedError()

        def reset_mock(self):
            self.take_called = False
            self.restore_called = False

    isolation_manager = IsolationManagerFailingAtRestoring(
        fixtures.isolation_manager.config_wrapper,
        fixtures.isolation_manager.receipt_capture,
        chain_snapshots=chain_snapshots,
    )
    isolation_context = isolation_manager.isolation(Scope.FUNCTION)
    next(isolation_context)  # Enter.
    # Snapshot works, we get this far.
    assert isolation_manager.take_called
    assert not isolation_manager.restore_called

    # At this point, it realized snapshotting is no-go.
    next(isolation_context, None)  # Exit.
    assert isolation_manager.restore_called

    isolation_manager.reset_mock()
    isolation_context = isolation_manager.isolation(Scope.FUNCTION)
    next(isolation_context)  # Enter again.

    # This time, snapshotting is NOT attempted.
    assert not isolation_manager.take_called
    assert not isolation_manager.restore_called


def test_isolation_supported_flag_set_after_successful_snapshot(fixtures):
    """
    Testing the unusual case where `.supported` was changed manually after
    a successful snapshot and before the restore attempt.
    """

    class CustomIsolationManager(IsolationManager):
        take_called = False
        restore_called = False

        def take_snapshot(self) -> Optional["SnapshotID"]:
            self.take_called = True
            return 123

        def restore(self, scope: Scope):
            self.restore_called = True

    isolation_manager = CustomIsolationManager(
        fixtures.isolation_manager.config_wrapper,
        fixtures.isolation_manager.receipt_capture,
    )
    isolation_context = isolation_manager.isolation(Scope.FUNCTION)
    next(isolation_context)  # Enter.
    assert isolation_manager.take_called
    assert not isolation_manager.restore_called

    # HACK: Change the flag manually to show it will avoid
    #   the restore.
    isolation_manager.supported = False
    isolation_manager.take_called = False  # Reset
    next(isolation_context, None)  # Exit.
    # Even though snapshotting worked, the flag was changed,
    # and so the restore never gets attempted.
    assert not isolation_manager.take_called



from typing import TYPE_CHECKING

import pytest
from eth_utils import to_hex
from rich.table import Table
from rich.tree import Tree

from ape.exceptions import ContractLogicError, OutOfGasError
from ape.utils import ManagerAccessMixin
from ape_ethereum.transactions import DynamicFeeTransaction, Receipt, TransactionStatusEnum

if TYPE_CHECKING:
    from ape.api import ReceiptAPI


@pytest.fixture
def deploy_receipt(vyper_contract_instance):
    return vyper_contract_instance.creation_metadata.receipt


@pytest.fixture
def invoke_receipt(vyper_contract_instance, owner):
    return vyper_contract_instance.setNumber(1, sender=owner)


@pytest.fixture
def trace_print_capture(mocker, chain):
    get_rich_console_patch_0 = mocker.patch("ape_ethereum.trace.get_rich_console")
    get_rich_console_patch_1 = mocker.patch("ape.managers.chain.get_rich_console")
    mock_console = mocker.MagicMock()
    get_rich_console_patch_0.return_value = mock_console
    get_rich_console_patch_1.return_value = mock_console
    return mock_console.print


def test_receipt_properties(chain, invoke_receipt):
    assert invoke_receipt.block_number == chain.blocks.head.number
    assert invoke_receipt.timestamp == chain.blocks.head.timestamp
    assert invoke_receipt.datetime == chain.blocks.head.datetime


def test_show_trace(trace_print_capture, invoke_receipt):
    invoke_receipt.show_trace()
    call_args = trace_print_capture.call_args_list

    # Trace title assertion.
    actual_title = call_args[0][0][0]
    assert actual_title == f"Call trace for [bold blue]'{invoke_receipt.txn_hash}'[/]"

    # The origin comes below the title.
    actual_origin = call_args[1][0][0]
    assert actual_origin == f"tx.origin=[#ff8c00]{invoke_receipt.sender}[/]"

    # Next are events.
    events_title = call_args[3][0][0]
    event_tree = call_args[4][0][0]
    event_label = str(event_tree.label)
    assert events_title == "Events emitted:"
    assert isinstance(event_tree, Tree)
    assert "[bright_green]NumberChange" in event_label
    assert 'dynData=[bright_magenta]"Dynamic"' in event_label
    assert "newNum=[bright_magenta]1" in event_label

    # Assert stuff about actual call-tree now.
    actual_calltree = call_args[6][0][0]
    assert isinstance(actual_calltree, Tree)
    label = f"{actual_calltree.label}"
    assert "VyperContract" in label
    assert "setNumber" in label
    assert f"[{invoke_receipt.gas_used} gas]" in label


def test_show_gas_report(trace_print_capture, invoke_receipt):
    invoke_receipt.show_gas_report()
    actual = trace_print_capture.call_args[0][0]
    assert isinstance(actual, Table)
    assert actual.title == "VyperContract Gas"


def test_show_events(trace_print_capture, invoke_receipt):
    invoke_receipt.show_events()
    actual = trace_print_capture.call_args[0][0]
    assert isinstance(actual, Tree)
    label = str(actual.label)
    # Formatted and enriched signature string.
    assert "[bright_green]NumberChange" in label
    assert 'dynData=[bright_magenta]"Dynamic"' in label
    assert "newNum=[bright_magenta]1" in label


def test_decode_logs(owner, contract_instance, assert_log_values):
    event_type = contract_instance.NumberChange

    # Invoke a transaction 3 times that generates 3 logs.
    receipt_0 = contract_instance.setNumber(1, sender=owner)
    receipt_1 = contract_instance.setNumber(2, sender=owner)
    receipt_2 = contract_instance.setNumber(3, sender=owner)

    def assert_receipt_logs(receipt: "ReceiptAPI", num: int):
        logs = receipt.decode_logs(event_type)
        assert len(logs) == 1
        assert_log_values(logs[0], num)
        assert receipt.timestamp == logs[0].timestamp

    assert_receipt_logs(receipt_0, 1)
    assert_receipt_logs(receipt_1, 2)
    assert_receipt_logs(receipt_2, 3)


def test_decode_logs_specify_abi(invoke_receipt, vyper_contract_instance):
    abi = vyper_contract_instance.NumberChange.abi
    logs = invoke_receipt.decode_logs(abi=abi)
    assert len(logs) == 1
    assert logs[0].newNum == 1
    assert logs[0].event_name == "NumberChange"
    assert logs[0].log_index == 0
    assert logs[0].transaction_index == 0


def test_decode_logs_specify_abi_as_event(
    mocker, invoke_receipt, vyper_contract_instance, eth_tester_provider
):
    spy = mocker.spy(eth_tester_provider.web3.eth, "get_logs")
    abi = vyper_contract_instance.NumberChange
    logs = invoke_receipt.decode_logs(abi=abi)
    assert len(logs) == 1
    assert logs[0].newNum == 1
    assert logs[0].event_name == "NumberChange"
    assert logs[0].log_index == 0
    assert logs[0].transaction_index == 0

    # Tests against a bug where the API was called unnecessarily
    assert spy.call_count == 0


def test_decode_logs_multiple_event_types(owner, contract_instance, assert_log_values):
    foo_happened = contract_instance.FooHappened
    bar_happened = contract_instance.BarHappened
    receipt = contract_instance.fooAndBar(sender=owner)
    logs = receipt.decode_logs([foo_happened, bar_happened])
    assert len(logs) == 2
    assert logs[0].foo == 0
    assert logs[1].bar == 1


def test_decode_logs_not_checksummed_addresses(owner, contract_instance, assert_log_values):
    """
    If an RPC returns non-checksummed logs with non-checksummed addresses,
    show we can still decode them.
    """
    start_number = contract_instance.myNumber()
    tx = contract_instance.setNumber(1, sender=owner)
    assert len(tx.logs) == 1

    # Hack to make the address lower-case.
    tx.logs[0]["address"] = tx.logs[0]["address"].lower()

    events = tx.decode_logs()
    assert len(events) == 1
    assert_log_values(events[0], 1, start_number)


def test_decode_logs_unspecified_abi_gets_all_logs(owner, contract_instance):
    receipt = contract_instance.fooAndBar(sender=owner)
    logs = receipt.decode_logs()  # Same as doing `receipt.events`
    assert len(logs) == 2
    assert logs[0].foo == 0
    assert logs[1].bar == 1


def test_events(owner, contract_instance, assert_log_values):
    receipt = contract_instance.setNumber(1, sender=owner)
    assert len(receipt.events) == 1
    assert_log_values(receipt.events[0], 1)


def test_events_with_ds_notes(ds_note_test_contract, owner):
    contract = ds_note_test_contract
    receipt = contract.test_0(sender=owner)
    assert len(receipt.events) == 1
    assert receipt.events[0].event_name == "foo"
    assert receipt.events[0].log_index == 0
    assert receipt.events[0].transaction_index == 0

    receipt = contract.test_1(sender=owner)
    assert len(receipt.events) == 1
    assert receipt.events[0].event_name == "foo"
    assert receipt.events[0].event_arguments == {"a": 1}
    assert receipt.events[0].log_index == 0
    assert receipt.events[0].transaction_index == 0

    receipt = contract.test_2(sender=owner)
    assert len(receipt.events) == 1
    assert receipt.events[0].event_name == "foo"
    assert receipt.events[0].event_arguments == {"a": 1, "b": 2}
    assert receipt.events[0].log_index == 0
    assert receipt.events[0].transaction_index == 0

    receipt = contract.test_3(sender=owner)
    assert len(receipt.events) == 1
    assert receipt.events[0].event_name == "foo"
    assert receipt.events[0].event_arguments == {"a": 1, "b": 2, "c": 3}
    assert receipt.events[0].log_index == 0
    assert receipt.events[0].transaction_index == 0


def test_get_failed_receipt(owner, vyper_contract_instance, eth_tester_provider):
    # Setting to '5' always fails.

    with pytest.raises(ContractLogicError) as err:
        vyper_contract_instance.setNumber(5, sender=owner, gas=100000)
        assert err.value.txn
        receipt = err.value.txn.receipt
        assert receipt
        assert receipt.failed


def test_receipt_raise_for_status_out_of_gas_error(mocker, ethereum):
    gas_limit = 100000
    txn = ethereum.create_transaction(
        chain_id=0,
        receiver="",
        sender="",
        gas_limit=gas_limit,
        nonce=0,
        value=0,
        data=b"",
        type=0,
        max_fee=None,
        max_priority_fee=None,
        required_confirmations=None,
    )
    receipt = Receipt(
        provider=mocker.MagicMock(),
        txn_hash="",
        gas_limit=gas_limit,
        gas_used=gas_limit,
        status=TransactionStatusEnum.FAILING,
        gas_price=0,
        block_number=0,
        transaction=txn,
    )
    with pytest.raises(OutOfGasError) as err:
        receipt.raise_for_status()

    assert err.value.txn == receipt


def test_receipt_chain_id(invoke_receipt, eth_tester_provider):
    assert invoke_receipt.chain_id == eth_tester_provider.chain_id


def test_track_coverage(deploy_receipt, mocker):
    """
    Show that deploy receipts are not tracked.
    """
    mock_runner = mocker.MagicMock()
    mock_tracker = mocker.MagicMock()
    mock_runner.coverage_tracker = mock_tracker
    original = ManagerAccessMixin._test_runner
    ManagerAccessMixin._test_runner = mock_runner

    deploy_receipt.track_coverage()

    assert mock_runner.track_coverage.call_count == 0
    ManagerAccessMixin._test_runner = original


def test_access_from_tx(deploy_receipt):
    actual = deploy_receipt.transaction.receipt
    assert actual == deploy_receipt


def test_transaction(owner, vyper_contract_instance):
    receipt = vyper_contract_instance.setNumber(1999, sender=owner)
    actual = receipt.transaction
    assert actual.sender == owner.address
    assert to_hex(actual.txn_hash) == receipt.txn_hash


def test_transaction_validated_from_dict(ethereum, owner, deploy_receipt):
    tx = ethereum.create_transaction(sender=owner.address, value=123, data=b"hello")
    tx_data = tx.model_dump()
    receipt_data = deploy_receipt.model_dump(by_alias=True)
    receipt_data["transaction"] = tx_data
    receipt = Receipt.model_validate(receipt_data)
    assert isinstance(receipt.transaction, DynamicFeeTransaction)
    assert receipt.transaction.sender == owner.address
    assert receipt.transaction.value == 123
    assert receipt.transaction.data == b"hello"


def test_return_value(owner, solidity_contract_instance):
    """
    ``.return_value`` still works when using EthTester provider!
    It works by using eth_call to get the result rather than
    tracing-RPCs.
    """
    solidity_contract_instance.loadArrays(sender=owner)
    receipt = solidity_contract_instance.getNestedArrayMixedDynamic.transact(sender=owner)
    actual = receipt.return_value
    assert len(actual) == 5
    assert actual[1][1] == [[0], [0, 1], [0, 1, 2]]



import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any

import click
from packaging.version import Version

from ape.cli.options import ape_cli_context, skip_confirmation_option
from ape.logging import logger

if TYPE_CHECKING:
    from ape.plugins._utils import PluginMetadata


@click.group(short_help="Manage ape plugins")
def cli():
    """
    Command-line helper for managing plugins.
    """


def plugins_argument():
    """
    An argument that is either the given list of plugins
    or plugins loaded from the local config file.
    """

    def load_from_file(ctx, file_path: Path) -> list["PluginMetadata"]:
        from ape.plugins._utils import PluginMetadata
        from ape.utils.misc import load_config

        if file_path.is_dir():
            name_options = (
                "ape-config.yaml",
                "ape-config.yml",
                "ape-config.json",
                "pyproject.toml",
            )
            for option in name_options:
                if (file_path / option).is_file():
                    file_path = file_path / option
                    break

        if file_path.is_file():
            config = load_config(file_path)
            if plugins := config.get("plugins"):
                return [PluginMetadata.model_validate(d) for d in plugins]

        ctx.obj.logger.warning(f"No plugins found at '{file_path}'.")
        return []

    def callback(ctx, param, value: tuple[str]):
        from ape.plugins._utils import PluginMetadata

        res = []
        if not value:
            ctx.obj.abort("You must give at least one requirement to install.")

        elif len(value) == 1:
            # User passed in a path to a file.
            file_path = Path(value[0]).expanduser().resolve()
            res = (
                load_from_file(ctx, file_path)
                if file_path.exists()
                else [PluginMetadata(name=v) for v in value[0].split(" ")]
            )

        else:
            res = [PluginMetadata(name=v) for v in value]

        return res

    return click.argument(
        "plugins",
        callback=callback,
        nargs=-1,
        metavar="PLUGIN-NAMES or path/to/project-dir",
    )


def upgrade_option(help: str = "", **kwargs):
    """
    A ``click.option`` for upgrading plugins (``--upgrade``).

    Args:
        help (str): CLI option help text. Defaults to ``""``.
    """

    return click.option("-U", "--upgrade", default=False, is_flag=True, help=help, **kwargs)


def validate_python_path(ctx, param, value):
    if value is None:
        return None

    path = Path(value).expanduser().resolve()
    if not path.exists():
        ctx.fail(f"Python path does not exist: {value}")

    return f"{path}"


def python_location_option(help: str = "", **kwargs):
    default_help = "Specify path to Python interpreter or virtualenv to use for installation"
    return click.option(
        "--python",
        type=str,
        default=None,
        callback=validate_python_path,
        help=help or default_help,
        **kwargs,
    )


def _display_all_callback(ctx, param, value):
    from ape.plugins._utils import PluginType

    return (
        (PluginType.CORE, PluginType.INSTALLED, PluginType.THIRD_PARTY, PluginType.AVAILABLE)
        if value
        else (PluginType.INSTALLED, PluginType.THIRD_PARTY)
    )


@cli.command(name="list", short_help="Display plugins")
@ape_cli_context()
@click.option(
    "-a",
    "--all",
    "to_display",
    default=False,
    is_flag=True,
    callback=_display_all_callback,
    help="Display all plugins installed and available (including Core)",
)
@click.option(
    "--format",
    "output_format",
    type=click.Choice(["default", "prefixed", "freeze"]),
    default="default",
)
@click.option("--exclude-version", is_flag=True, help="Do not include plugin versions in output")
def _list(cli_ctx, to_display, output_format, exclude_version):
    from ape.plugins._utils import PluginMetadataList, PluginType

    include_available = PluginType.AVAILABLE in to_display
    metadata = PluginMetadataList.load(cli_ctx.plugin_manager, include_available=include_available)
    if output := metadata.to_str(
        include=to_display, include_version=not exclude_version, output_format=output_format
    ):
        click.echo(output)
        if not metadata.installed and not metadata.third_party:
            click.echo("No plugins installed (besides core plugins).")

    else:
        click.echo("No plugins installed.")


@cli.command()
@ape_cli_context()
@plugins_argument()
@skip_confirmation_option("Don't ask for confirmation to install the plugins")
@upgrade_option(help="Upgrade the plugin to the newest available version")
@python_location_option(help="Specify the virtualenv location")
def install(
    cli_ctx, plugins: list["PluginMetadata"], skip_confirmation: bool, upgrade: bool, python: str
):
    """Install plugins"""

    failures_occurred = False

    # Track the operations until the end. This way, if validation
    # fails on one, we can error-out before installing anything.
    install_list: list[dict[str, Any]] = []

    for plugin in plugins:
        result = plugin._prepare_install(
            upgrade=upgrade, skip_confirmation=skip_confirmation, python_location=python
        )
        if result:
            install_list.append(result)
        else:
            failures_occurred = True

    # NOTE: Be *extremely careful* with `subprocess.call`, as it modifies the user's
    #       installed packages, to potentially catastrophic results
    for op in install_list:
        if not op:
            continue

        handler = op["result_handler"]
        call_result = subprocess.call(op["args"])
        if "version_before" in op:
            success = not handler.handle_upgrade_result(call_result, op["version_before"])
            failures_occurred = not failures_occurred and success

        else:
            success = not handler.handle_install_result(call_result)
            failures_occurred = not failures_occurred and success

    if failures_occurred:
        sys.exit(1)


@cli.command()
@ape_cli_context()
@plugins_argument()
@skip_confirmation_option("Don't ask for confirmation to install the plugins")
@python_location_option(help="Specify the virtualenv location")
def uninstall(cli_ctx, plugins, skip_confirmation, python: str):
    """Uninstall plugins"""
    from ape.plugins._utils import ModifyPluginResultHandler

    failures_occurred = False
    did_warn_about_version = False
    for plugin in plugins:
        if plugin.version is not None and not did_warn_about_version:
            cli_ctx.logger.warning("Specifying a version when uninstalling is not necessary.")
            did_warn_about_version = True

        result_handler = ModifyPluginResultHandler(plugin)

        # if plugin is installed but not a 2nd class. It must be a third party
        if plugin.is_installed and not plugin.is_available:
            cli_ctx.logger.warning(
                f"Plugin '{plugin.name}' is installed but not in available plugins. "
                f"Please uninstall outside of Ape."
            )
            failures_occurred = True
            continue

        # check for installed check the config.yaml
        elif not plugin.is_installed:
            cli_ctx.logger.warning(f"Plugin '{plugin.name}' is not installed.")
            failures_occurred = True
            continue

        # if plugin is installed and 2nd class. We should uninstall it
        if plugin.is_installed and (
            skip_confirmation or click.confirm(f"Remove plugin '{plugin}'?")
        ):
            cli_ctx.logger.info(f"Uninstalling '{plugin.name}'...")
            arguments = plugin._get_uninstall_args(python_location=python)

            # NOTE: Be *extremely careful* with this command, as it modifies the user's
            #       installed packages, to potentially catastrophic results
            # NOTE: This is not abstracted into another function *on purpose*
            result = subprocess.call(arguments)
            failures_occurred = not result_handler.handle_uninstall_result(result)

    if failures_occurred:
        sys.exit(1)


@cli.command()
def update():
    """
    Update Ape and all plugins to the next version
    """
    from ape.plugins._utils import ape_version

    _change_version(ape_version.next_version_range)


def _version_callback(ctx, param, value):
    obj = Version(value)
    version_str = f"0.{obj.minor}.0" if obj.major == 0 else f"{obj.major}.0.0"
    return f"=={version_str}"


@cli.command()
@click.argument("version", callback=_version_callback)
def change_version(version):
    """
    Change ape and all plugins version
    """

    _change_version(version)


def _install(name, spec, exit_on_fail: bool = True) -> int:
    """
    Helper function to install or update a Python package using pip.

    Args:
      name (str): The package name.
      spec (str): Version specifier, e.g., '==1.0.0', '>=1.0.0', etc.
      exit_on_fail (bool): Set to ``False`` to not exit on fail.

    Returns:
        The process return-code.
    """
    from ape.plugins._utils import PIP_COMMAND

    arguments = [*PIP_COMMAND, "install", f"{name}{spec}", "--quiet"]

    # Run the installation process and capture output for error checking
    completed_process = subprocess.run(
        arguments,
        capture_output=True,
        text=True,  # Output as string
        check=False,  # Allow manual error handling
    )

    # Check for installation errors
    if completed_process.returncode != 0:
        message = f"Failed to install/update {name}"
        if completed_process.stdout:
            message += f": {completed_process.stdout}"
        if completed_process.stderr:
            message += f": {completed_process.stderr}"

        logger.error(message)
        if exit_on_fail:
            sys.exit(completed_process.returncode)
    else:
        logger.info(f"Successfully installed/updated {name}")

    return completed_process.returncode


def _change_version(spec: str):
    # Update all the plugins.
    # This will also update core Ape.
    # NOTE: It is possible plugins may depend on each other and may update in
    #   an order causing some error codes to pop-up, so we ignore those for now.
    from ape.plugins._utils import get_plugin_dists

    plugin_retcode = 0
    for plugin in get_plugin_dists():
        logger.info(f"Updating {plugin} ...")
        name = plugin.split("=")[0].strip()
        retcode = _install(name, spec, exit_on_fail=False)
        if retcode != 0:
            plugin_retcode = retcode
        # else: errors logged in _install separately

    # This check is for verifying the update and shouldn't actually do anything.
    logger.info("Updating Ape core ...")
    ape_retcode = _install("eth-ape", spec)
    if ape_retcode == 0 and plugin_retcode == 0:
        prefix = "Ape"
        if plugin_retcode == 0:
            prefix = f"{prefix} and plugins"

        logger.success(f"{prefix} have successfully upgraded.")
    # else: _install logs errors already.
    exit_code = ape_retcode if ape_retcode != 0 else plugin_retcode
    sys.exit(exit_code)



import pytest
from eth_typing import ChecksumAddress
from hypothesis import given
from hypothesis import strategies as st

from ape.exceptions import ConversionError
from ape_ethereum._converters import ETHER_UNITS

TEN_THOUSAND_ETHER_IN_WEI = 10_000_000_000_000_000_000_000


@pytest.mark.fuzzing
@given(
    value=st.decimals(
        min_value=-(2**255),
        max_value=2**256 - 1,
        allow_infinity=False,
        allow_nan=False,
    ),
    unit=st.sampled_from(list(ETHER_UNITS.keys())),
)
def test_ether_conversions(value, unit, convert):
    currency_str = f"{value} {unit}"
    actual = convert(currency_str, int)
    expected = int(value * ETHER_UNITS[unit])
    assert actual == expected
    # Also show can compare directly with str.
    assert actual == currency_str


def test_bad_type(convert):
    with pytest.raises(ConversionError) as err:
        convert("something", float)

    expected = (
        "Type '<class 'float'>' must be one of [AddressType, bytes, int, Decimal, bool, str]."
    )
    assert str(err.value) == expected


def test_no_registered_converter(convert):
    with pytest.raises(ConversionError) as err:
        convert("something", ChecksumAddress)

    assert str(err.value) == "No conversion registered to handle 'something'."


@pytest.mark.parametrize("sep", (",", "_"))
def test_separaters(convert, sep):
    """
    Show that separates, such as commands and underscores, are OK
    in currency-string values, e.g. "10,000 ETH" is valid.
    """
    currency_str = f"10{sep}000 ETHER"
    actual = convert(currency_str, int)
    expected = TEN_THOUSAND_ETHER_IN_WEI
    assert actual == expected



import doctest

extensions = ["sphinx_ape"]

# -- Doctest configuration --

doctest_default_flags = (
    0
    | doctest.DONT_ACCEPT_TRUE_FOR_1
    | doctest.ELLIPSIS
    | doctest.IGNORE_EXCEPTION_DETAIL
    | doctest.NORMALIZE_WHITESPACE
)

doctest_global_setup = """
from ape import *
"""



import os
from collections.abc import Iterable, Iterator
from enum import Enum
from functools import cached_property
from pathlib import Path
from typing import Any, TypeVar, cast

import yaml
from ethpm_types import PackageManifest, PackageMeta, Source
from pydantic import ConfigDict, Field, ValidationError, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from ape.exceptions import ConfigError
from ape.logging import logger
from ape.types.address import AddressType
from ape.utils.abi import CalldataRepr
from ape.utils.basemodel import (
    ExtraAttributesMixin,
    ExtraModelAttributes,
    ManagerAccessMixin,
    _assert_not_ipython_check,
    get_attribute_with_extras,
    only_raise_attribute_error,
)
from ape.utils.misc import load_config
from ape.utils.os import clean_path

ConfigItemType = TypeVar("ConfigItemType")


class ConfigEnum(str, Enum):
    """
    A configuration `Enum <https://docs.python.org/3/library/enum.html>`__ type.
    Use this to limit the values of a config item, such as colors ``"RED"``, ``"BLUE"``,
    ``"GREEN"``, rather than any arbitrary ``str``.

    Usage example::

            class MyEnum(ConfigEnum):
                FOO = "FOO"
                BAR = "BAR"


            class MyConfig(PluginConfig):
                my_enum: MyEnum


            model = MyConfig(my_enum="FOO")

    """


def _find_config_yaml_files(base_path: Path) -> list[Path]:
    """
    Find all ape config file in the given path.
    """
    found: list[Path] = []
    if (base_path / "ape-config.yaml").is_file():
        found.append(base_path / "ape-config.yaml")
    if (base_path / "ape-config.yml").is_file():
        found.append(base_path / "ape-config.yml")

    return found


class PluginConfig(BaseSettings):
    """
    A base plugin configuration class. Each plugin that includes
    a config API must register a subclass of this class.
    """

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_")

    @classmethod
    def from_overrides(
        cls, overrides: dict, plugin_name: str | None = None, project_path: Path | None = None
    ) -> "PluginConfig":
        default_values = cls().model_dump()

        def update(root: dict, value_map: dict):
            for key, val in value_map.items():
                if isinstance(val, dict) and key in root and isinstance(root[key], dict):
                    root[key] = update(root[key], val)
                else:
                    root[key] = val

            return root

        data = update(default_values, overrides)
        try:
            return cls.model_validate(data)
        except ValidationError as err:
            plugin_name = plugin_name or cls.__name__.replace("Config", "").lower()
            if problems := cls._find_plugin_config_problems(
                err, plugin_name, project_path=project_path
            ):
                raise ConfigError(problems) from err
            else:
                raise ConfigError(str(err)) from err

    @classmethod
    def _find_plugin_config_problems(
        cls, err: ValidationError, plugin_name: str, project_path: Path | None = None
    ) -> str | None:
        # Attempt showing line-nos for failed plugin config validation.
        # This is trickier than root-level data since by this time, we
        # no longer are aware of which files are responsible for which config.
        ape = ManagerAccessMixin

        # First, try checking the root config file ALONE. It is important to do
        # w/o any data from the project-level config to isolate the source of the problem.
        raw_global_data = ape.config_manager.global_config.model_dump(by_alias=True)
        if plugin_name in raw_global_data:
            try:
                cls.model_validate(raw_global_data[plugin_name])
            except Exception:
                if problems := cls._find_plugin_config_problems_from_file(
                    err, ape.config_manager.DATA_FOLDER
                ):
                    return problems

        # No issues found with global; try the local project.
        # NOTE: No need to isolate project-data w/o root-data because we have already
        #  determined root-level data is OK.
        project_path = project_path or ape.local_project.path
        if problems := cls._find_plugin_config_problems_from_file(err, project_path):
            return problems

        return None

    @classmethod
    def _find_plugin_config_problems_from_file(
        cls, err: ValidationError, base_path: Path
    ) -> str | None:
        cfg_files = _find_config_yaml_files(base_path)
        for cfg_file in cfg_files:
            if problems := _get_problem_with_config(err.errors(), cfg_file):
                return problems

        return None

    @only_raise_attribute_error
    def __getattr__(self, attr_name: str) -> Any:
        _assert_not_ipython_check(attr_name)

        # Allow hyphens in plugin config files.
        attr_name = attr_name.replace("-", "_")
        extra = self.__pydantic_extra__ or {}
        if attr_name in extra:
            return extra[attr_name]

        return super().__getattribute__(attr_name)

    def __getitem__(self, item: str) -> Any:
        extra = self.__pydantic_extra__ or {}
        if item in self.__dict__:
            return self.__dict__[item]

        elif item in extra:
            return extra[item]

        raise KeyError(f"'{item}' not in config.")

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__ or key in (self.__pydantic_extra__ or {})

    def __str__(self) -> str:
        data = self.model_dump(
            mode="json",
            by_alias=True,
            exclude_none=True,
            exclude_unset=True,
            exclude_defaults=True,
        )
        return yaml.safe_dump(data)

    def get(self, key: str, default: ConfigItemType | None = None) -> ConfigItemType:
        extra: dict = self.__pydantic_extra__ or {}
        return self.__dict__.get(key, extra.get(key, default))  # type: ignore


class GenericConfig(ConfigDict):
    """
    The default class used when no specialized class is used.
    """


class DeploymentConfig(PluginConfig):
    """
    Add 'deployments' to your config.
    """

    address: AddressType
    """
    The address of the deployment.
    """

    contract_type: str
    """
    The contract type name reference
    (must be a contract in the project).
    """


def _get_problem_with_config(errors: list, path: Path) -> str | None:
    # Attempt to find line numbers in the config matching.
    cfg_content = Source(content=path.read_text(encoding="utf8")).content
    if not cfg_content:
        # Likely won't happen?
        return None

    err_map: dict = {}
    for error in errors:
        if not (location := error.get("loc")):
            continue

        lineno = None
        loc_idx = 0
        depth = len(location)
        list_counter = 0
        for no, line in cfg_content.items():
            loc = location[loc_idx]
            clean_line = line.lstrip()

            if isinstance(loc, int):
                if not clean_line.startswith("- "):
                    # Not a list item.
                    continue

                elif list_counter != loc:
                    # Still search for index.
                    list_counter += 1
                    continue

                # If we get here, we have found the index.
                list_counter = 0

            elif not clean_line.startswith(f"{loc}:"):
                continue

            # Look for the next location up the tree.
            loc_idx += 1

            # Even if we don't find the full path for some reason
            # (perhaps it's missing), we still have the last lineno noticed.
            lineno = no

            if loc_idx < depth:
                continue

            # Found.
            break

        if lineno is not None:
            err_map[lineno] = error
        # else: we looped through the whole file and didn't find anything.

    if not err_map:
        # raise ValidationError as-is (no line numbers found for some reason).
        return None

    error_strs: list[str] = []
    for line_no, cfg_err in err_map.items():
        sub_message = cfg_err.get("msg", cfg_err)
        line_before = line_no - 1 if line_no > 1 else None
        line_after = line_no + 1 if line_no < len(cfg_content) else None
        lines = []
        if line_before is not None:
            lines.append(f"   {line_before}: {cfg_content[line_before]}")
        lines.append(f"-->{line_no}: {cfg_content[line_no]}")
        if line_after is not None:
            lines.append(f"   {line_after}: {cfg_content[line_after]}")

        file_preview = "\n".join(lines)
        sub_message = f"{sub_message}\n{file_preview}\n"
        error_strs.append(sub_message)

    # NOTE: Using reversed because later pydantic errors
    #   appear earlier in the list.
    final_msg = "\n".join(reversed(error_strs)).strip()
    return f"'{clean_path(path)}' is invalid!\n{final_msg}"


class DisplayConfig(PluginConfig):
    """
    Configure display settings in Ape.
    """

    calldata: CalldataRepr = CalldataRepr.abridged
    """
    Dictates how the calldata displays when signing transactions.
    """


class ApeConfig(ExtraAttributesMixin, BaseSettings, ManagerAccessMixin):
    """
    The top-level config.
    """

    def __init__(self, *args, **kwargs):
        project_path = kwargs.get("project")
        super().__init__(*args, **kwargs)
        # NOTE: Cannot reference `self` at all until after super init.
        self._project_path = project_path

    contracts_folder: str | None = None
    """
    The path to the folder containing the contract source files.
    **NOTE**: Non-absolute paths are relative to the project-root.
    If not set, defaults to deducing the contracts folder.
    When deducing, Ape first tries ``"contracts"``, but if
    that folder does not exist, Ape tries to find a folder with
    contracts.
    """

    default_ecosystem: str = "ethereum"
    """
    The default ecosystem to use in Ape.
    """

    dependencies: list[dict] = []
    """
    Project dependency declarations.
    Note: The actual dependency classes are decoded later.
    """

    deployment_data: dict[str, dict[str, list[DeploymentConfig]]] = Field(
        default_factory=dict, alias="deployments"
    )
    """
    Data for deployed contracts from the project.
    """

    display: DisplayConfig = DisplayConfig()
    """
    Configure display settings in Ape.
    """

    interfaces_folder: str = "interfaces"
    """
    The path to the project's interfaces.
    """

    meta: PackageMeta = PackageMeta()
    """
    Metadata about the active project as per EIP
    https://eips.ethereum.org/EIPS/eip-2678#the-package-meta-object
    """

    name: str = ""
    """
    The name of the project.
    """

    base_path: str | None = None
    """
    Use this when the project's base-path is not the
    root of the project.
    """

    request_headers: dict = {}
    """
    Extra request headers for all HTTP requests.
    """

    version: str = ""
    """
    The version of the project.
    """

    # NOTE: Plugin configs are technically "extras".
    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_")

    @model_validator(mode="before")
    @classmethod
    def validate_model(cls, model):
        model = model or {}
        fixed_model: dict = {}
        for key, val in model.items():
            # Allows hyphens to work anywhere where underscores are.
            fixed_model[key.replace("-", "_")] = val

        if project := fixed_model.pop("project", None):
            # Resolve local dependencies so relative paths don't cause
            # problems when moving the project around (as happens in local
            # dependencies).
            fixed_deps = []
            dependencies_list = fixed_model.get("dependencies", [])
            if not isinstance(dependencies_list, Iterable):
                raise ConfigError(
                    "Expecting dependencies: to be iterable. "
                    f"Received: {type(dependencies_list).__name__}"
                )

            for dep in fixed_model.get("dependencies", []):
                if not isinstance(dep, dict):
                    raise ConfigError(
                        f"Expecting mapping for dependency. Received: {type(dep).__name__}."
                    )

                fixed_dep = {**dep}
                if "project" not in fixed_dep:
                    fixed_dep["project"] = project
                # else: we might be told to use a different project.
                #   when decoding dependencies, the project is mostly used for
                #  stuff like resolving paths. If this is already set,
                #  this is likely a dependency of a dependency.

                fixed_deps.append(fixed_dep)

            if fixed_deps:
                fixed_model["dependencies"] = fixed_deps

        # field: contracts_folder: Handle if given Path object.
        if "contracts_folder" in fixed_model and isinstance(fixed_model["contracts_folder"], Path):
            fixed_model["contracts_folder"] = str(fixed_model["contracts_folder"])

        return fixed_model

    @cached_property
    def deployments(self) -> dict[str, dict[str, list[DeploymentConfig]]]:
        # Lazily validated.
        for ecosystem_name, ecosystem_deploys in self.deployment_data.items():
            if ecosystem_name not in self.network_manager.ecosystems:
                raise ConfigError(f"Invalid ecosystem '{ecosystem_name}' in deployments config.")

            ecosystem = self.network_manager.ecosystems[ecosystem_name]
            for network_name, network_deploys in ecosystem_deploys.items():
                if network_name not in ecosystem.networks:
                    raise ConfigError(
                        f"Invalid network '{ecosystem_name}:{network_name}' in deployments config."
                    )

        return self.deployment_data

    def __ape_extra_attributes__(self) -> Iterator[ExtraModelAttributes]:
        # This allows `config.my_plugin` to work.
        yield ExtraModelAttributes(
            name="plugin_configs",
            attributes=lambda n: self.get_config(n.replace("-", "_")),
            include_getitem=True,
        )

    @classmethod
    def validate_file(cls, path: Path, **overrides) -> "ApeConfig":
        """
        Create an ApeConfig class using the given path.
        Supports both pyproject.toml and ape-config.[.yml|.yaml|.json] files.

        Raises:
            :class:`~ape.exceptions.ConfigError`: When given an unknown file type
              or the data is invalid.

        Args:
            path (Path): The path to the file.
            **overrides: Config overrides.

        Returns:
            :class:`~ape.api.config.ApeConfig`
        """
        file_config = load_config(path)
        file_config = {k.replace("-", "_"): v for k, v in file_config.items()}
        overrides = {k.replace("-", "_"): v for k, v in overrides.items()}
        data = {**file_config, **overrides}

        # NOTE: We are including the project path here to assist
        #  in relative-path resolution, such as for local dependencies.
        #  You can use relative paths in local dependencies in your
        #  ape-config.yaml file but Ape needs to know the source to be
        #  relative from.
        data["project"] = path.parent

        try:
            return cls.model_validate(data)
        except ValidationError as err:
            if path.suffix == ".json":
                # TODO: Support JSON configs here.
                raise ConfigError(f"{err}") from err

            if final_msg := _get_problem_with_config(err.errors(), path):
                raise ConfigError(final_msg)
            else:
                raise ConfigError(str(err)) from err

    @classmethod
    def from_manifest(cls, manifest: PackageManifest, **overrides) -> "ApeConfig":
        return cls.model_validate(
            {
                **_get_compile_configs_from_manifest(manifest),
                **_get_dependency_configs_from_manifest(manifest),
                **overrides,
            }
        )

    @property
    def _plugin_configs(self) -> dict:
        # NOTE: Ensure a dict exists so we can add to it
        #  and have it persist.
        self.__pydantic_extra__ = self.__pydantic_extra__ or {}
        return self.__pydantic_extra__

    def __repr__(self) -> str:
        return "<ape-config.yaml>"

    def __str__(self) -> str:
        data = self.model_dump(
            mode="json",
            by_alias=True,
            exclude_none=True,
            exclude_unset=True,
            exclude_defaults=True,
        )
        return yaml.safe_dump(data)

    @only_raise_attribute_error
    def __getattr__(self, attr_name: str) -> Any:
        return get_attribute_with_extras(self, attr_name)

    def __getitem__(self, name: str) -> Any:
        return self.__getattr__(name)

    def __contains__(self, item):
        # Always return True to handle lazy loading plugins.
        return True

    def model_dump(self, *args, **kwargs):
        res = super().model_dump(*args, **kwargs)
        # TODO: For some reason underscore prefixed kwargs
        #  still show up even though Pydantic says they
        #  shouldn't. Figure out why.
        return {k: v for k, v in res.items() if not k.startswith("_")}

    def get(self, name: str) -> Any | None:
        return self.__getattr__(name)

    def get_config(self, plugin_name: str) -> PluginConfig:
        name = plugin_name.replace("-", "_")
        return (
            self.get_plugin_config(name)
            or self.get_custom_ecosystem_config(name)
            or self.get_unknown_config(name)
        )

    def get_plugin_config(self, name: str) -> PluginConfig | None:
        name = name.replace("-", "_")
        cfg = self._plugin_configs.get(name, {})
        if cfg and not isinstance(cfg, dict):
            # Already decoded.
            return cfg

        for plugin_name, config_class in self._get_config_plugin_classes():
            cls: type[PluginConfig] = config_class  # type: ignore
            if plugin_name.replace("-", "_") != name:
                continue

            if cls != ConfigDict:
                # NOTE: Will raise if improperly provided keys
                config = cls.from_overrides(
                    cfg, plugin_name=plugin_name, project_path=self._project_path
                )
            else:
                # NOTE: Just use it directly as a dict if `ConfigDict` is passed
                config = cfg

            self._plugin_configs[name] = config
            return config

        return None

    def _get_config_plugin_classes(self):
        # NOTE: Abstracted for easily mocking in tests.
        return self.plugin_manager.config_class

    def get_custom_ecosystem_config(self, name: str) -> PluginConfig | None:
        name = name.replace("-", "_")
        if not (networks := self.get_plugin_config("networks")):
            # Shouldn't happen.
            return None

        for network in networks.custom:
            if name not in (network.ecosystem, network.ecosystem.replace("-", "_")):
                continue

            # Check if has a cached override.
            override = self._plugin_configs.get(name, {})
            if not isinstance(override, dict):
                return override

            # Custom network found.
            from ape_ethereum import EthereumConfig

            ethereum = cast(EthereumConfig, self.get_plugin_config("ethereum"))
            return ethereum.from_overrides(
                override, plugin_name=name, project_path=self._project_path
            )

        return None

    def get_unknown_config(self, name: str) -> PluginConfig:
        # This happens when a plugin is not installed but still configured.
        result = (self.__pydantic_extra__ or {}).get(name, PluginConfig())
        if isinstance(result, dict):
            return PluginConfig.from_overrides(
                result, plugin_name=name, project_path=self._project_path
            )

        return result

    def write_to_disk(self, destination: Path, replace: bool = False):
        """
        Write this config to a file.

        Args:
            destination (Path): The path to write to.
            replace (bool): Set to ``True`` to overwrite the file if it exists.
        """
        if destination.exists() and not replace:
            raise ValueError(f"Destination {destination} exists.")
        elif replace:
            destination.unlink(missing_ok=True)

        if destination.suffix in (".yml", ".yaml"):
            destination.parent.mkdir(parents=True, exist_ok=True)
            with open(destination, "x") as file:
                data = self.model_dump(by_alias=True, mode="json")
                yaml.safe_dump(data, file)

        elif destination.suffix == ".json":
            destination.write_text(self.model_dump_json(by_alias=True), encoding="utf8")

        else:
            raise ConfigError(f"Unsupported destination file type '{destination}'.")


def _get_compile_configs_from_manifest(manifest: PackageManifest) -> dict[str, dict]:
    configs: dict[str, dict] = {}
    for compiler in [x for x in manifest.compilers or [] if x.settings]:
        name = compiler.name.strip().lower()
        compiler_data = {}
        settings = compiler.settings or {}
        remapping_list = []
        for remapping in settings.get("remappings") or []:
            parts = remapping.split("=")
            key = parts[0]
            link = parts[1]
            if link.startswith(f".cache{os.path.sep}"):
                link = os.path.sep.join(link.split(f".cache{os.path.sep}"))[1:]

            new_entry = f"{key}={link}"
            remapping_list.append(new_entry)

        if remapping_list:
            compiler_data["import_remapping"] = remapping_list

        if "evm_version" in settings:
            compiler_data["evm_version"] = settings["evm_version"]

        if compiler_data:
            configs[name] = compiler_data

    return configs


def _get_dependency_configs_from_manifest(manifest: PackageManifest) -> dict:
    dependencies_config: list[dict] = []
    dependencies = manifest.dependencies or {}
    for package_name, uri in dependencies.items():
        if "://" not in str(uri) and hasattr(uri, "scheme"):
            uri_str = f"{uri.scheme}://{uri}"
        else:
            uri_str = str(uri)

        dependency: dict = {"name": str(package_name)}
        if uri_str.startswith("https://github.com/") and "releases/tag" in uri_str:
            # 'https:', '', 'github.com', org, repo, 'releases', 'tag', version
            # Fails with ValueError if not matching
            try:
                _, _, _, org, repo, _, _, version = uri_str.split("/")
            except ValueError:
                raise ConfigError("")

            dependency["github"] = f"{org}/{repo}"

            # If version fails, the dependency system will automatically try `ref`.
            dependency["version"] = version

        elif uri_str.startswith("file://"):
            dependency["local"] = uri_str.replace("file://", "")

        else:
            logger.error(f"Manifest URI {uri_str} not a supported dependency.")
            continue

        dependencies_config.append(dependency)

    return {"dependencies": dependencies_config} if dependencies_config else {}



import time
from collections.abc import Callable
from random import randint

import requests
from requests.models import CaseInsensitiveDict

from ape.exceptions import ProviderError, ProviderNotConnectedError
from ape.logging import logger
from ape.utils.misc import __version__, _python_version

USER_AGENT: str = f"Ape/{__version__} (Python/{_python_version})"


def allow_disconnected(fn: Callable):
    """
    A decorator that instead of raising :class:`~ape.exceptions.ProviderNotConnectedError`
    warns and returns ``None``.

    Usage example::

        from typing import Optional
        from ape.types import SnapshotID
        from ape.utils import return_none_when_disconnected


        @allow_disconnected
        def try_snapshot(self) -> "SnapshotID | None":
            return self.chain.snapshot()

    """

    def inner(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except ProviderNotConnectedError:
            logger.warning("Provider is not connected.")
            return None

    return inner


def stream_response(download_url: str, progress_bar_description: str = "Downloading") -> bytes:
    """
    Download HTTP content by streaming and returning the bytes.
    Progress bar will be displayed in the CLI.

    Args:
        download_url (str): String to get files to download.
        progress_bar_description (str): Downloading word.

    Returns:
        bytes: Content in bytes to show the progress.
    """
    # Lazy import for performance
    from tqdm import tqdm  # type: ignore[import-untyped]

    response = requests.get(download_url, stream=True)
    response.raise_for_status()

    total_size = int(response.headers.get("content-length", 0))
    progress_bar = tqdm(total=total_size, unit="iB", unit_scale=True, leave=False)
    progress_bar.set_description(progress_bar_description)
    content = b""
    for data in response.iter_content(1024, decode_unicode=True):
        progress_bar.update(len(data))
        content += data

    progress_bar.close()
    return content


class RPCHeaders(CaseInsensitiveDict):
    """
    A dict-like data-structure for HTTP-headers.
    It is case-insensitive and appends user-agent strings
    rather than overrides.
    """

    def __setitem__(self, key, value):
        if key.lower() != "user-agent" or not self.__contains__("user-agent"):
            return super().__setitem__(key, value)

        # Handle appending the user-agent (without replacing).
        existing_user_agent = self.__getitem__("user-agent")
        parts = [a.strip() for a in value.split(" ")]
        new_parts = []
        for part in parts:
            if part in existing_user_agent:
                # Already added.
                continue
            else:
                new_parts.append(part)

        if new_user_agent := " ".join(new_parts):
            super().__setitem__(key, f"{existing_user_agent} {new_user_agent}")


def request_with_retry(
    func: Callable,
    min_retry_delay: int = 1_000,
    retry_backoff_factor: int = 2,
    max_retry_delay: int = 30_000,
    max_retries: int = 10,
    retry_jitter: int = 250,
    is_rate_limit: Callable[[Exception], bool] | None = None,
):
    """
    Make a request with 429/rate-limit retry logic.

    Args:
        func (Callable): The function to run with rate-limit handling logic.
        min_retry_delay (int): The amount of milliseconds to wait before
          retrying the request. Defaults to ``1_000`` (one second).
        retry_backoff_factor (int): The multiplier applied to the retry delay
          after each failed attempt. Defaults to ``2``.
        max_retry_delay (int): The maximum length of the retry delay.
          Defaults to ``30_000`` (30 seconds).
        max_retries (int): The maximum number of retries.
          Defaults to ``10``.
        retry_jitter (int): A random number of milliseconds up to this limit
          is added to each retry delay. Defaults to ``250`` milliseconds.
        is_rate_limit (Callable[[Exception], bool] | None): A custom handler
          for detecting rate-limits. Defaults to checking for a 429 status
          code on an HTTPError.
    """
    if not is_rate_limit:
        # Use default checker.
        def checker(err: Exception) -> bool:
            return isinstance(err, requests.HTTPError) and err.response.status_code == 429

        is_rate_limit = checker

    for attempt in range(max_retries):
        try:
            return func()
        except Exception as err:
            if not is_rate_limit(err):
                # It was not a rate limit error. Raise whatever exception it is.
                raise

            else:
                # We were rate-limited. Invoke retry/backoff logic.
                logger.warning("Request was rate-limited. Backing-off and then retrying...")
                retry_interval = min(
                    max_retry_delay, min_retry_delay * retry_backoff_factor**attempt
                )
                delay = retry_interval + randint(0, retry_jitter)
                time.sleep(delay / 1000)
                continue

    # If we get here, we over-waited. Raise custom exception.
    raise ProviderError(f"Rate limit retry-mechanism exceeded after '{max_retries}' attempts.")



from typing import TYPE_CHECKING

from evm_trace.gas import merge_reports

from ape.utils.basemodel import ManagerAccessMixin
from ape.utils.trace import _exclude_gas, parse_gas_table

if TYPE_CHECKING:
    from ethpm_types.abi import MethodABI
    from ethpm_types.source import ContractSource

    from ape.api.trace import TraceAPI
    from ape.pytest.config import ConfigWrapper
    from ape.types.address import AddressType
    from ape.types.trace import ContractFunctionPath, GasReport


class GasTracker(ManagerAccessMixin):
    """
    Class for tracking gas-used per method called in the
    contracts in your test suite.
    """

    def __init__(self, config_wrapper: "ConfigWrapper"):
        self.config_wrapper = config_wrapper
        self.session_gas_report: GasReport | None = None

    @property
    def enabled(self) -> bool:
        return self.config_wrapper.track_gas

    @property
    def gas_exclusions(self) -> list["ContractFunctionPath"]:
        return self.config_wrapper.gas_exclusions

    def show_session_gas(self) -> bool:
        if not self.session_gas_report:
            return False

        tables = parse_gas_table(self.session_gas_report)
        self.chain_manager._reports.echo(*tables)
        return True

    def append_gas(self, trace: "TraceAPI", contract_address: "AddressType"):
        contract_type = self.chain_manager.contracts.get(contract_address)
        if not contract_type:
            # Skip unknown contracts.
            return

        report = trace.get_gas_report(exclude=self.gas_exclusions)
        self._merge(report)

    def append_toplevel_gas(self, contract: "ContractSource", method: "MethodABI", gas_cost: int):
        exclusions = self.gas_exclusions or []
        if (contract_id := contract.contract_type.name) and not _exclude_gas(
            exclusions, contract_id, method.selector
        ):
            self._merge({contract_id: {method.selector: [gas_cost]}})

    def _merge(self, report: dict):
        session_report = self.session_gas_report or {}
        self.session_gas_report = merge_reports(session_report, report)



"""
TODO: In 0.9, move this module to `ape.types`.
"""

import inspect
import json
from abc import ABC
from collections.abc import Callable, Iterator, Mapping, Sequence
from importlib import import_module
from pathlib import Path
from sys import getrecursionlimit
from typing import TYPE_CHECKING, Any, ClassVar

from ethpm_types import BaseModel as EthpmTypesBaseModel
from pydantic import BaseModel as RootBaseModel
from pydantic import ConfigDict

from ape.exceptions import (
    ApeAttributeError,
    ApeIndexError,
    ProviderNotConnectedError,
)
from ape.logging import logger
from ape.utils.misc import log_instead_of_fail, raises_not_implemented

if TYPE_CHECKING:
    from pydantic.main import Model

    from ape.api.providers import ProviderAPI
    from ape.managers.accounts import AccountManager
    from ape.managers.chain import ChainManager
    from ape.managers.compilers import CompilerManager
    from ape.managers.config import ConfigManager
    from ape.managers.converters import ConversionManager
    from ape.managers.networks import NetworkManager
    from ape.managers.plugins import PluginManager
    from ape.managers.project import DependencyManager, ProjectManager
    from ape.managers.query import QueryManager
    from ape.pytest.runners import PytestApeRunner


class classproperty:
    def __init__(self, fn: Callable):
        self.fn = fn

    def __get__(self, obj, owner) -> Any:
        return self.fn(owner)


class manager_access(property):
    _cache = None

    def __init__(self, fn):
        self.fn = fn

    def __get__(self, obj, owner) -> Any:  # type: ignore[override]
        if self._cache is None:
            self._cache = self.fn(owner)

        return self._cache


class _RecursionChecker:
    # A helper for preventing the recursion errors
    # that happen in custom __getattr__ methods.

    def __init__(self):
        self.THRESHOLD: int = getrecursionlimit()
        self.getattr_checking: dict[str, int] = {}
        self.getattr_errors: dict[str, Exception] = {}

    @log_instead_of_fail(default="<_RecursionChecker>")
    def __repr__(self) -> str:
        return repr(self.getattr_checking)

    def check(self, name: str) -> bool:
        return (self.getattr_checking.get(name, 0) or 0) >= self.THRESHOLD

    def add(self, name: str):
        if name in self.getattr_checking:
            self.getattr_checking[name] += 1
        else:
            self.getattr_checking[name] = 1

    def reset(self, name: str):
        self.getattr_checking.pop(name, None)
        self.getattr_errors.pop(name, None)


_recursion_checker = _RecursionChecker()


# TODO: Delete in 0.9 (deprecated & no longer used anywhere)
class injected_before_use(property):
    """
    Injected properties are injected class variables that must be set before use.

    **NOTE**: do not appear in a Pydantic model's set of properties.
    """

    def __get__(self, *args):
        arg_strs = []
        for argument in args:
            try:
                arg_str = str(argument)
            except Exception as err:
                logger.debug(f"Failed calling __str__. Exception: {err}")
                arg_strs.append("<?>")
                continue

            arg_strs.append(arg_str)

        error_message = "Value not set"
        if arg_strs:
            error_message = f"{error_message} (arguments={', '.join(arg_strs)})"

        error_message = f"{error_message}. Please inject this property before calling."
        raise ValueError(error_message)


def only_raise_attribute_error(fn: Callable) -> Any:
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except AttributeError:
            raise  # Don't modify or log attr errors.
        except Exception as err:
            # Wrap the exception in AttributeError
            logger.log_debug_stack_trace()
            raise ApeAttributeError(f"{err}", base_err=err) from err

    return wrapper


class ManagerAccessMixin:
    """
    A mixin for accessing Ape's manager at the class level.

    Usage example:

        from ape.utils import ManagerAccessMixin

        class MyClass(ManagerAccessMixin):
            def my_function(self):
                accounts = self.account_manager  # And so on!
    """

    _test_runner: ClassVar["PytestApeRunner | None"] = None

    @manager_access
    def account_manager(cls) -> "AccountManager":
        """
        The :class:`~ape.managers.accounts.AccountManager`.
        """
        accounts = import_module("ape.managers.accounts")
        return accounts.AccountManager()

    @manager_access
    def chain_manager(cls) -> "ChainManager":
        """
        The :class:`~ape.managers.chain.ChainManager`.
        """
        chain = import_module("ape.managers.chain")
        return chain.ChainManager()

    @manager_access
    def compiler_manager(cls) -> "CompilerManager":
        """
        The :class:`~ape.managers.compilers.CompilerManager`.
        """
        compilers = import_module("ape.managers.compilers")
        return compilers.CompilerManager()

    @manager_access
    def config_manager(cls) -> "ConfigManager":
        """
        The :class:`~ape.managers.config.ConfigManager`.
        """
        config = import_module("ape.managers.config")
        return config.ConfigManager()

    @manager_access
    def conversion_manager(cls) -> "ConversionManager":
        """
        The :class:`~ape.managers.converters.ConversionManager`.
        """
        converters = import_module("ape.managers.converters")
        return converters.ConversionManager()

    @manager_access
    def local_project(cls) -> "ProjectManager":
        """
        A :class:`~ape.managers.project.ProjectManager` pointed
        at the current-working directory.
        """
        project = import_module("ape.managers.project")
        return project.ProjectManager(Path.cwd())

    @manager_access
    def network_manager(cls) -> "NetworkManager":
        """
        The :class:`~ape.managers.networks.NetworkManager`.
        """
        networks = import_module("ape.managers.networks")
        return networks.NetworkManager()

    @manager_access
    def plugin_manager(cls) -> "PluginManager":
        """
        The :class:`~ape.managers.plugins.PluginManager`.
        """
        plugins = import_module("ape.managers.plugins")
        return plugins.PluginManager()

    @classproperty
    def Project(cls) -> type["ProjectManager"]:
        """
        The ``Project`` factory class for creating
        other local-projects.
        """
        project = import_module("ape.managers.project")
        return project.ProjectManager

    @manager_access
    def query_manager(cls) -> "QueryManager":
        """
        The :class:`~ape.managers.query.QueryManager`.
        """
        query = import_module("ape.managers.query")
        return query.QueryManager()

    @classproperty
    def provider(cls) -> "ProviderAPI":
        """
        The current active provider if connected to one.

        Raises:
            :class:`~ape.exceptions.ProviderNotConnectedError`: When there is
            no active provider at runtime.

        Returns:
            :class:`~ape.api.providers.ProviderAPI`
        """
        if provider := cls.network_manager.active_provider:
            return provider

        raise ProviderNotConnectedError()

    @classproperty
    def dependency_manager(cls) -> "DependencyManager":
        # We make this available for more intuitive access to
        # global dependencies, which any project has access to.
        return cls.local_project.dependencies


class BaseInterface(ManagerAccessMixin, ABC):
    """
    Abstract class that has manager access.
    """


def _get_alt(name: str) -> str | None:
    alt = None
    if ("-" not in name and "_" not in name) or ("-" in name and "_" in name):
        alt = None
    elif "-" in name:
        alt = name.replace("-", "_")
    elif "_" in name:
        alt = name.replace("_", "-")

    return alt


class _AttrLookup(dict):
    """
    Used when given extra attributes via a callback
    that takes the attribute name.
    """

    def __init__(
        self,
        callback: Callable[
            [
                str,
            ],
            None,
        ],
    ):
        self._callback = callback

    def __contains__(self, item) -> bool:
        return self._callback(item) is not None

    @only_raise_attribute_error
    def __getattr__(self, item):
        res = self._callback(item)
        if res is None:
            # attr-lookups cannot return None!
            raise AttributeError(item)

        return res

    def __getitem__(self, item):
        return self._callback(item)

    def get(self, item):  # type: ignore[override]
        return self._callback(item)


class ExtraModelAttributes(EthpmTypesBaseModel):
    """
    A class for defining extra model attributes.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str
    """
    The name of the attributes. This is important
    in instances such as when an attribute is missing,
    we can show a more accurate exception message.
    """

    attributes: Any | Callable[[], Any] | Callable[[str], Any]
    """The attributes. The following types are supported:

    1. A model or dictionary to lookup attributes.
    2. A callable with no arguments, for lazily evaluating a model or dictionary
       for lookup.
    3. A callable with a single argument that is the attribute name. This style
       of lookup cannot be used for optionals.
    """

    include_getattr: bool = True
    """Whether to use these in ``__getattr__``."""

    include_getitem: bool = False
    """Whether to use these in ``__getitem__``."""

    additional_error_message: str | None = None
    """
    An additional error message to include at the end of
    the normal IndexError message.
    """

    def __repr__(self) -> str:
        try:
            return f"<ExtraAttributes '{self.name}'>"
        except Exception:
            # Disallow exceptions in __repr__
            return "<ExtraModelAttributes>"

    def __contains__(self, name: Any) -> bool:
        attrs = self._attrs()
        try:
            name = str(name)
        except Exception:
            return False

        if name in attrs or hasattr(attrs, name):
            return True

        elif alt := _get_alt(name):
            return alt in attrs

        return False

    def get(self, name: str) -> Any | None:
        """
        Get an attribute.

        Args:
            name (str): The name of the attribute.

        Returns:
            Optional[Any]: The attribute if it exists, else ``None``.
        """

        res = self._get(name)
        if res is not None:
            return res

        if alt := _get_alt(name):
            res = self._get(alt)
            if res is not None:
                return res

        return None

    def _get(self, name: str) -> Any | None:
        attrs = self._attrs()
        return attrs.get(name) if hasattr(attrs, "get") else getattr(attrs, name, None)

    def _attrs(self) -> Any:
        if not isinstance(self.attributes, Callable):  # type: ignore
            # Dict or model that can do a lookup.
            return self.attributes

        signature = inspect.signature(self.attributes)
        if len(signature.parameters) == 0:
            # Lazy-eval dict.
            return self.attributes()  # type: ignore

        # Callable lookup via name.
        return _AttrLookup(self.attributes)  # type: ignore


class BaseModel(EthpmTypesBaseModel):
    """
    An ape-pydantic BaseModel.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def model_copy(
        self: "Model",
        *,
        update: Mapping[str, Any] | None = None,
        deep: bool = False,
        cache_clear: Sequence[str] | None = None,
    ) -> "Model":
        result = super().model_copy(update=update, deep=deep)

        # Clear @cached_properties
        for cached_item in cache_clear or []:
            if cached_item in result.__dict__:
                del result.__dict__[cached_item]

        return result

    @raises_not_implemented
    def _repr_mimebundle_(self, include=None, exclude=None):
        # This works better than AttributeError for Ape.
        pass

    @raises_not_implemented
    def _ipython_display_(self, include=None, exclude=None):
        # This works better than AttributeError for Ape.
        pass


def _assert_not_ipython_check(key):
    # Perf: IPython expects AttributeError here.
    if isinstance(key, str) and key == "_ipython_canary_method_should_not_exist_":
        raise AttributeError()


class ExtraAttributesMixin:
    """
    A mixin to use on models that provide ``ExtraModelAttributes``.
    **NOTE**: Must come _before_ your base-model class in subclass tuple to function.
    """

    def __ape_extra_attributes__(self) -> Iterator[ExtraModelAttributes]:
        """
        Override this method to supply extra attributes
        to a model in Ape; this allow more properties
        to be available when invoking ``__getattr__``.

        Returns:
            Iterator[:class:`~ape.utils.basemodel.ExtraModelAttributes`]: A
            series of instances defining extra model attributes.
        """
        return iter(())

    @only_raise_attribute_error
    def __getattr__(self, name: str) -> Any:
        """
        An overridden ``__getattr__`` implementation that takes into
        account :meth:`~ape.utils.basemodel.ExtraAttributesMixin.__ape_extra_attributes__`.
        """
        _assert_not_ipython_check(name)
        private_attrs = (self.__pydantic_private__ or {}) if isinstance(self, RootBaseModel) else {}
        if name in private_attrs:
            _recursion_checker.reset(name)
            return private_attrs[name]

        return get_attribute_with_extras(self, name)

    def __getitem__(self, name: Any) -> Any:
        # For __getitem__, we first try the extra (unlike `__getattr__`).
        return get_item_with_extras(self, name)


def get_attribute_with_extras(obj: Any, name: str, coerce_attr_error: bool = True) -> Any:
    _assert_not_ipython_check(name)
    if _recursion_checker.check(name):
        # Prevent segfaults.
        # First, attempt to get real error.
        message = f"Failed trying to get {name}"
        if real_error := _recursion_checker.getattr_errors.get(name):
            message = f"{message}. {real_error}"

        _recursion_checker.reset(name)
        raise AttributeError(message)

    _recursion_checker.add(name)

    res = None

    if not isinstance(obj, ExtraAttributesMixin):
        name = getattr(type(obj), "__name__", "obj")
        raise AttributeError(f"{name} must use the '{ExtraAttributesMixin.__name__}' mixin'")

    try:
        res = super(ExtraAttributesMixin, obj).__getattribute__(name)
    except AttributeError as base_attr_err:
        _recursion_checker.getattr_errors[name] = base_attr_err

    if res is not None:
        _recursion_checker.reset(name)
        return res

    if name.startswith("__") and name.endswith("__"):
        # Don't seek double-dunderized definitions from extras.
        raise AttributeError(name)

    # NOTE: Do not check extras within the error handler to avoid
    #   errors occurring within an exception handler (Python shows that differently).
    extras_checked = set()
    for ape_extra in obj.__ape_extra_attributes__():
        if not ape_extra.include_getattr:
            continue

        extras_checked.add(ape_extra.name)
        try:
            if name in ape_extra:
                # Attribute was found in one of the supplied
                # extra attributes mappings.
                result = ape_extra.get(name)
                # NOTE: Don't reset until _after_ we have the result.
                _recursion_checker.reset(name)
                return result

        except Exception as err:
            _recursion_checker.reset(name)
            raise ApeAttributeError(f"{name} - {err}", base_err=err) from err

    # The error message mentions the alternative mappings,
    # such as a contract-type map.
    base_err = None
    if name in _recursion_checker.getattr_errors:
        # There was an error getting the value. Show that.
        base_err = _recursion_checker.getattr_errors[name]
        message = str(base_err)
    else:
        message = f"'{obj!r}' has no attribute '{name}'"

    if extras_checked:
        extras_str = ", ".join(sorted(extras_checked))
        suffix = f"Also checked extra(s) '{extras_str}'"
        if suffix not in message:
            if message and message[-1] not in (".", "?", "!"):
                message = f"{message}."

            message = f"{message} {suffix}"

    _recursion_checker.reset(name)
    if message and message[-1] not in (".", "?", "!"):
        message = f"{message}."

    if base_err and not coerce_attr_error:
        raise base_err

    # Coerce whatever error to automatically be an AttributeError
    # (required for __getattr__ or must handle independently).
    attr_err = ApeAttributeError(message)
    if base_err:
        raise attr_err from base_err
    else:
        raise attr_err


def get_item_with_extras(obj: Any, name: str) -> Any:
    # For __getitem__, we first try the extra (unlike `__getattr__`).
    extras_checked = set()
    additional_error_messages = {}
    for extra in obj.__ape_extra_attributes__():
        if not extra.include_getitem:
            continue

        if name in extra:
            return extra.get(name)

        extras_checked.add(extra.name)

        if extra.additional_error_message:
            additional_error_messages[extra.name] = extra.additional_error_message

    # NOTE: If extras were supplied, the user was expecting it to be
    #   there (unlike __getattr__).
    if extras_checked:
        prefix = f"Unable to find '{name}' in"
        if not additional_error_messages:
            extras_str = ", ".join(extras_checked)
            message = f"{prefix} any of '{extras_str}'."

        else:
            # The class is including additional error messages for the IndexError.
            message = ""
            for extra_checked in extras_checked:
                additional_message = additional_error_messages.get(extra_checked)
                suffix = f" {additional_message}" if additional_message else ""
                sub_message = f"{prefix} '{extra_checked}'.{suffix}"
                message = f"{message}\n{sub_message}" if message else sub_message

        raise ApeIndexError(message)

    # The user did not supply any extra __getitem__ attributes.
    # Do what you would have normally done.
    return super(ExtraAttributesMixin, obj).__getitem__(name)  # type: ignore


class BaseInterfaceModel(BaseInterface, BaseModel):
    """
    An abstract base-class with manager access on a pydantic base model.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __dir__(self) -> list[str]:
        """
        **NOTE**: Should integrate options in IPython tab-completion.
        https://ipython.readthedocs.io/en/stable/config/integrating.html
        """
        # Filter out protected/private members
        return [member for member in super().__dir__() if not member.startswith("_")]


class DiskCacheableModel(BaseModel):
    """
    A model with extra utilities for caching to disk.
    """

    def __init__(self, *args, **kwargs):
        path = kwargs.pop("path", None)
        super().__init__(*args, **kwargs)
        self._path = path

    def model_read_file(self, path: Path | None = None) -> dict:
        """
        Get the file's raw data. This is different from ``model_dump()`` because it
        reads directly from the file without validation.
        """
        path = self._get_path(path=path)
        return self._model_read_file(path)

    @classmethod
    def _model_read_file(cls, path: Path) -> dict:
        """
        Get the file's raw data. This is different from ``model_dump()`` because it
        reads directly from the file without validation.
        """
        if json_str := path.read_text(encoding="utf8") if path.is_file() else "":
            return json.loads(json_str)

        return {}

    def model_dump_file(self, path: Path | None = None, **kwargs):
        """
        Save this model to disk.

        Args:
            path (Path | None): Optionally provide the path now
              if one wasn't declared at init time. If given a directory,
              saves the file in that dir with the name of class with a
              .json suffix.
            **kwargs: Extra kwargs to pass to ``.model_dump_json()``.
        """
        path = self._get_path(path=path)
        json_str = self.model_dump_json(**kwargs)
        path.unlink(missing_ok=True)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(json_str)

    @classmethod
    def model_validate_file(cls, path: Path, **kwargs):
        """
        Validate a file.

        Args:
            path (Path): Optionally provide the path now
              if one wasn't declared at init time.
            **kwargs: Extra kwargs to pass to ``.model_validate_json()``.
        """
        data = cls._model_read_file(path)
        model = cls.model_validate(data, **kwargs)
        model._path = path
        return model

    def _get_path(self, path: Path | None = None) -> Path:
        if save_path := (path or self._path):
            return save_path

        elif save_path.is_dir():
            name = self.__class__.__name__ or "Model"
            return save_path / f"{name}.json"

        raise ValueError("Unknown path for caching.")



from collections.abc import Iterator
from functools import singledispatchmethod

from ape.api.query import ContractCreation, ContractCreationQuery, QueryAPI, QueryType
from ape.exceptions import APINotImplementedError, ProviderError, QueryEngineError
from ape.types.address import AddressType


class EthereumQueryProvider(QueryAPI):
    """
    Implements more advanced queries specific to Ethereum clients.
    """

    def __init__(self):
        self.supports_contract_creation = None  # will be set after we try for the first time

    @singledispatchmethod
    def estimate_query(self, query: QueryType) -> int | None:  # type: ignore[override]
        return None

    @singledispatchmethod
    def perform_query(self, query: QueryType) -> Iterator:  # type: ignore[override]
        raise QueryEngineError(f"Cannot handle '{type(query)}'.")

    @estimate_query.register
    def estimate_contract_creation_query(self, query: ContractCreationQuery) -> int | None:
        # NOTE: Extremely expensive query, involves binary search of all blocks in a chain
        #       Very loose estimate of 5s per transaction for this query.
        if self.supports_contract_creation is False:
            return None
        return 5000

    @perform_query.register
    def perform_contract_creation_query(
        self, query: ContractCreationQuery
    ) -> Iterator[ContractCreation]:
        """
        Find when a contract was deployed using binary search and block tracing.
        """
        # skip the search if there is still no code at address at head
        if not self.chain_manager.get_code(query.contract):
            return None

        def find_creation_block(lo, hi):
            # perform a binary search to find the block when the contract was deployed.
            # takes log2(height), doesn't work with contracts that have been reinit.
            while hi - lo > 1:
                mid = (lo + hi) // 2
                code = self.chain_manager.get_code(query.contract, block_id=mid)
                if not code:
                    lo = mid
                else:
                    hi = mid

            if self.chain_manager.get_code(query.contract, block_id=hi):
                return hi

            return None

        try:
            block = find_creation_block(0, self.chain_manager.blocks.height)
        except ProviderError:
            self.supports_contract_creation = False
            return None

        # iterate over block transaction traces to find the deployment call
        # this method also supports contracts created by factories
        try:
            if "geth" in self.provider.client_version.lower():
                yield from self._find_creation_in_block_via_geth(block, query.contract)
            else:
                yield from self._find_creation_in_block_via_parity(block, query.contract)
        except (ProviderError, APINotImplementedError):
            self.supports_contract_creation = False
            return None

        self.supports_contract_creation = True

    def _find_creation_in_block_via_parity(self, block, contract_address):
        # NOTE requires `trace_` namespace
        traces = self.provider.make_request("trace_replayBlockTransactions", [block, ["trace"]])

        for tx in traces:
            for trace in tx["trace"]:
                if (
                    "error" not in trace
                    and trace["type"] == "create"
                    and trace["result"]["address"] == contract_address.lower()
                ):
                    receipt = self.chain_manager.get_receipt(tx["transactionHash"])
                    creator = self.conversion_manager.convert(trace["action"]["from"], AddressType)
                    yield ContractCreation(
                        txn_hash=tx["transactionHash"],
                        block=block,
                        deployer=receipt.sender,
                        factory=creator if creator != receipt.sender else None,
                    )

    def _find_creation_in_block_via_geth(self, block, contract_address):
        # NOTE requires `debug_` namespace
        traces = self.provider.make_request(
            "debug_traceBlockByNumber", [hex(block), {"tracer": "callTracer"}]
        )

        def flatten(call):
            if call["type"] in ["CREATE", "CREATE2"]:
                yield call["from"], call["to"]

            if "error" in call or "calls" not in call:
                return

            for sub in call["calls"]:
                if sub["type"] in ["CREATE", "CREATE2"]:
                    yield sub["from"], sub["to"]
                else:
                    yield from flatten(sub)

        for tx in traces:
            call = tx["result"]
            sender = call["from"]
            for factory, contract in flatten(call):
                if contract == contract_address.lower():
                    yield ContractCreation(
                        txn_hash=tx["txHash"],
                        block=block,
                        deployer=self.conversion_manager.convert(sender, AddressType),
                        factory=(
                            self.conversion_manager.convert(factory, AddressType)
                            if factory != sender
                            else None
                        ),
                    )



import re

import pytest

from ape.pytest.contextmanagers import RevertsContextManager as reverts


def test_revert_info_context(owner, reverts_contract_instance):
    """
    Shows no two revert info objects are the same instance.
    """

    rev = reverts()
    with rev as rev0:
        reverts_contract_instance.revertStrings(0, sender=owner)
    with rev as rev1:
        reverts_contract_instance.revertStrings(1, sender=owner)

    assert rev0.value.revert_message != rev1.value.revert_message


def test_no_args(owner, reverts_contract_instance):
    """
    Test catching transaction reverts without asserting on error messages.
    """
    with reverts():
        reverts_contract_instance.revertStrings(0, sender=owner)


def test_revert(owner, reverts_contract_instance):
    """
    Test matching a revert message with a supplied message.
    """
    with reverts("zero"):
        reverts_contract_instance.revertStrings(0, sender=owner)


def test_revert_pattern(owner, reverts_contract_instance):
    """
    Test matching a revert message with a supplied pattern.
    """
    with reverts(re.compile(r"ze\w+")):
        reverts_contract_instance.revertStrings(0, sender=owner)


def test_revert_error(error_contract, not_owner):
    """
    Test matching a revert custom Solidity error.
    """
    with reverts(error_contract.Unauthorized):
        error_contract.withdraw(sender=not_owner)


def test_revert_abi(error_contract, not_owner):
    """
    Test matching a revert custom Solidity error using an ABI.
    """
    with reverts(error_contract.Unauthorized.abi):
        error_contract.withdraw(sender=not_owner)


def test_revert_error_from_container(project, error_contract, not_owner):
    """
    Test matching a revert custom Solidity error using the container instead
    of an instance, so the specific contract instance is not checked, only the
    ABI is. This is required for proper deploy-txn checks.
    """
    with reverts(project.HasError.Unauthorized):
        error_contract.withdraw(sender=not_owner)


def test_revert_error_from_container_with_expected_values(project, error_contract, not_owner):
    """
    Test matching a revert custom Solidity error using the container instead
    of an instance, so the specific contract instance is not checked, only the
    ABI is. This is required for proper deploy-txn checks.
    """
    with reverts(project.HasError.Unauthorized, error_inputs={"addr": not_owner.address}):
        error_contract.withdraw(sender=not_owner)


def test_revert_unexpected_error(error_contract, not_owner):
    """
    Test when given a different error type than what was raised.
    """
    expected = "Expected error 'OtherError' but was 'Unauthorized'"
    with pytest.raises(AssertionError, match=expected):
        with reverts(error_contract.OtherError):
            error_contract.withdraw(sender=not_owner)


def test_revert_error_inputs(error_contract, not_owner):
    """
    Test matching a revert custom Solidity error with inputs.
    """
    with reverts(error_contract.Unauthorized, addr=not_owner.address, counter=123):
        error_contract.withdraw(sender=not_owner)


def test_revert_error_unexpected_inputs(error_contract, owner, not_owner):
    """
    Test matching a revert custom Solidity error with unexpected inputs.
    """
    expected = (
        rf"Expected input 'addr' to be '{owner.address}' but was '{not_owner.address}'\."
        r"\nExpected input 'counter' to be '321' but was '123'\."
    )
    with pytest.raises(AssertionError, match=expected):
        with reverts(error_contract.Unauthorized, addr=owner.address, counter=321):
            error_contract.withdraw(sender=not_owner)


def test_revert_fails(owner, reverts_contract_instance):
    """
    Test that ``AssertionError`` is raised if the supplied revert reason does not match the actual
    revert reason.
    """
    with pytest.raises(AssertionError):
        with reverts("one"):
            reverts_contract_instance.revertStrings(0, sender=owner)


def test_revert_pattern_fails(owner, reverts_contract_instance):
    """
    Test that ``AssertionError`` is raised if the actual revert reason does not match the supplied
    revert pattern.
    """
    with pytest.raises(AssertionError):
        with reverts(re.compile(r"[^zero]+")):
            reverts_contract_instance.revertStrings(0, sender=owner)


def test_revert_partial_fails(owner, reverts_contract_instance):
    """
    Test that ``AssertionError`` is raised if the supplied revert reason does not match the actual
    revert reason exactly.
    """
    with pytest.raises(AssertionError):
        with reverts("ze"):
            reverts_contract_instance.revertStrings(0, sender=owner)



import sys
import time
from abc import abstractmethod
from collections.abc import Iterator
from datetime import datetime as datetime_type
from functools import cached_property
from typing import IO, TYPE_CHECKING, Any, NoReturn

from eth_pydantic_types import HexBytes, HexStr
from eth_utils import humanize_hexstr, is_hex, to_hex, to_int
from pydantic import ConfigDict, field_validator
from pydantic.fields import Field
from tqdm import tqdm  # type: ignore

from ape.exceptions import (
    NetworkError,
    ProviderNotConnectedError,
    SignatureError,
    TransactionError,
    TransactionNotFoundError,
)
from ape.logging import logger
from ape.types.address import AddressType
from ape.types.basic import HexInt
from ape.types.gas import AutoGasLimit
from ape.types.signatures import TransactionSignature
from ape.utils.basemodel import BaseInterfaceModel, ExtraAttributesMixin, ExtraModelAttributes
from ape.utils.misc import log_instead_of_fail, raises_not_implemented
from ape.utils.trace import prettify_function

if TYPE_CHECKING:
    from ethpm_types.abi import EventABI, MethodABI

    from ape.api.explorers import ExplorerAPI
    from ape.api.providers import BlockAPI
    from ape.api.trace import TraceAPI
    from ape.contracts import ContractEvent
    from ape.types.events import ContractLogContainer
    from ape.types.trace import SourceTraceback
    from ape.utils.abi import CalldataRepr


class TransactionAPI(BaseInterfaceModel):
    """
    An API class representing a transaction.
    Ecosystem plugins implement one or more of transaction APIs
    depending on which schemas they permit,
    such as typed-transactions from `EIP-1559 <https://eips.ethereum.org/EIPS/eip-1559>`__.
    """

    chain_id: HexInt | None = Field(default=0, alias="chainId")
    receiver: AddressType | None = Field(default=None, alias="to")
    sender: AddressType | None = Field(default=None, alias="from")
    gas_limit: HexInt | None = Field(default=None, alias="gas")
    nonce: HexInt | None = None  # NOTE: `None` only to denote using default behavior
    value: HexInt = 0
    data: HexBytes = HexBytes("")
    type: HexInt
    max_fee: HexInt | None = None
    max_priority_fee: HexInt | None = None

    # If left as None, will get set to the network's default required confirmations.
    required_confirmations: HexInt | None = Field(default=None, exclude=True)

    signature: TransactionSignature | None = Field(default=None, exclude=True)

    model_config = ConfigDict(populate_by_name=True)

    def __init__(self, *args, **kwargs):
        raise_on_revert = kwargs.pop("raise_on_revert", True)
        super().__init__(*args, **kwargs)
        self._raise_on_revert = raise_on_revert

    @field_validator("gas_limit", mode="before")
    @classmethod
    def validate_gas_limit(cls, value):
        if value is None:
            value = (
                cls.network_manager.active_provider.network.gas_limit
                if cls.network_manager.connected
                else 0
            )

        if value == "auto" or isinstance(value, AutoGasLimit):
            return None  # Delegate to `ProviderAPI.estimate_gas_cost`

        elif value == "max":
            if not cls.network_manager.connected:
                raise NetworkError("Must be connected to use 'max'.")

            return cls.network_manager.active_provider.max_gas

        elif isinstance(value, str) and is_hex(value):
            return to_int(hexstr=value)

        elif isinstance(value, str) and value.isnumeric():
            return to_int(value)

        return value

    @property
    def gas(self) -> int | None:
        """
        Alias for ``.gas_limit``.
        """
        return self.gas_limit

    @gas.setter
    def gas(self, value):
        self.gas_limit = self.validate_gas_limit(value)

    @property
    def raise_on_revert(self) -> bool:
        """
        ``True`` means VM-reverts should raise exceptions.
        ``False`` allows getting failed receipts.
        """
        return self._raise_on_revert

    @raise_on_revert.setter
    def raise_on_revert(self, value):
        self._raise_on_revert = value

    @property
    def total_transfer_value(self) -> int:
        """
        The total amount of WEI that a transaction could use.
        Useful for determining if an account balance can afford
        to submit the transaction.
        """
        if self.max_fee is None:
            raise TransactionError("`self.max_fee` must not be None.")

        return self.value + self.max_fee

    @property
    @abstractmethod
    def txn_hash(self) -> HexBytes:
        """
        The calculated hash of the transaction.
        """

    # TODO: In 0.9, simply rename txn_hash to hash.
    @property
    def hash(self) -> HexBytes:
        """
        Alias for ``self.txn_hash``.
        """
        return self.txn_hash

    @property
    def receipt(self) -> "ReceiptAPI | None":
        """
        This transaction's associated published receipt, if it exists.
        """
        try:
            txn_hash = to_hex(self.txn_hash)
        except SignatureError:
            return None

        try:
            return self.chain_manager.get_receipt(txn_hash)
        except (TransactionNotFoundError, ProviderNotConnectedError):
            return None

    @property
    def trace(self) -> "TraceAPI":
        """
        The transaction trace. Only works if this transaction was published
        and you are using a provider that support tracing.

        Raises:
            :class:`~ape.exceptions.APINotImplementedError`: When using a provider
              that does not support tracing.
        """
        return self.provider.get_transaction_trace(to_hex(self.txn_hash))

    @cached_property
    def _calldata_repr(self) -> "CalldataRepr":
        return self.local_project.config.display.calldata

    @abstractmethod
    def serialize_transaction(self) -> bytes:
        """
        Serialize the transaction
        """

    @log_instead_of_fail(default="<TransactionAPI>")
    def __repr__(self) -> str:
        # NOTE: Using JSON mode for style.
        data = self.model_dump(mode="json")
        params = ", ".join(f"{k}={v}" for k, v in data.items())
        cls_name = getattr(type(self), "__name__", TransactionAPI.__name__)
        return f"<{cls_name} {params}>"

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, calldata_repr: "CalldataRepr | None" = None) -> str:
        """
        Get the stringified representation of the transaction.

        Args:
            calldata_repr (:class:`~ape.types.abi.CalldataRepr` | None): Pass "full"
              to see the full calldata. Defaults to the value from the config.

        Returns:
            str
        """
        data = self.model_dump(mode="json")  # JSON mode used for style purposes.

        calldata_repr = calldata_repr or self._calldata_repr
        data["data"] = self._get_calldata_repr_str(calldata_repr=calldata_repr)

        params = "\n  ".join(f"{k}: {v}" for k, v in data.items())
        cls_name = getattr(type(self), "__name__", TransactionAPI.__name__)
        tx_str = f"{cls_name}:\n  {params}"

        # Decode the actual call so the user can see the function.
        if decoded := self._decoded_call():
            tx_str = f"{tx_str}\n\n\t{decoded}"

        return tx_str

    def _get_calldata_repr_str(self, calldata_repr: "CalldataRepr") -> str:
        calldata = HexBytes(self.data)

        # Elide the transaction calldata for abridged representations if the length exceeds 8
        # (4 bytes for function selector and trailing 4 bytes).
        return (
            calldata[:4].to_0x_hex() + "..." + calldata[-4:].hex()
            if calldata_repr == "abridged" and len(calldata) > 8
            else calldata.to_0x_hex()
        )

    def _decoded_call(self) -> str | None:
        if not self.receiver:
            return "constructor()"

        if not (contract_type := self.chain_manager.contracts.get(self.receiver)):
            # Unknown.
            return None

        try:
            abi = contract_type.methods[HexBytes(self.data)[:4]]
        except KeyError:
            return None

        ecosystem = (
            self.provider.network.ecosystem
            if self.network_manager.active_provider
            else self.network_manager.ethereum
        )
        decoded_calldata = ecosystem.decode_calldata(abi, HexBytes(self.data)[4:])

        # NOTE: There is no actual returndata yet, but we can show the type.
        return_types = [t.canonical_type for t in abi.outputs]
        if len(return_types) == 1:
            return_types = return_types[0]

        return prettify_function(
            abi.name or "",
            decoded_calldata,
            returndata=return_types,
            contract=contract_type.name or humanize_hexstr(self.receiver),
            is_create=self.receiver is None,
            depth=4,
        )


class ConfirmationsProgressBar:
    """
    A progress bar tracking the confirmations of a transaction.
    """

    def __init__(self, confirmations: int):
        self._req_confs = confirmations
        self._bar = tqdm(range(confirmations))
        self._confs = 0

    def __enter__(self):
        self._update_bar(0)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._bar.close()

    @property
    def confs(self) -> int:
        """
        The number of confirmations that have occurred.

        Returns:
            int: The total number of confirmations that have occurred.
        """
        return self._confs

    @confs.setter
    def confs(self, new_value):
        if new_value == self._confs:
            return

        diff = new_value - self._confs
        self._confs = new_value
        self._update_bar(diff)

    def _update_bar(self, amount: int):
        self._set_description()
        self._bar.update(amount)
        self._bar.refresh()

    def _set_description(self):
        self._bar.set_description(f"Confirmations ({self._confs}/{self._req_confs})")


class ReceiptAPI(ExtraAttributesMixin, BaseInterfaceModel):
    """
    An abstract class to represent a transaction receipt. The receipt
    contains information about the transaction, such as the status
    and required confirmations.

    **NOTE**: Use a ``required_confirmations`` of ``0`` in your transaction
    to not wait for confirmations.

    Get a receipt by making transactions in ``ape``, such as interacting with
    a :class:`ape.contracts.base.ContractInstance`.
    """

    contract_address: AddressType | None = None
    block_number: HexInt
    gas_used: HexInt
    logs: list[dict] = []
    status: HexInt
    txn_hash: HexStr
    transaction: TransactionAPI
    _error: TransactionError | None = None

    @log_instead_of_fail(default="<ReceiptAPI>")
    def __repr__(self) -> str:
        cls_name = getattr(self.__class__, "__name__", ReceiptAPI.__name__)
        return f"<{cls_name} {self.txn_hash}>"

    def __ape_extra_attributes__(self) -> Iterator[ExtraModelAttributes]:
        yield ExtraModelAttributes(name="transaction", attributes=lambda: vars(self.transaction))

    @field_validator("transaction", mode="before")
    @classmethod
    def _validate_transaction(cls, value):
        if not isinstance(value, dict):
            # Already a `TransactionAPI`.
            return value

        # Attempt to create a transaction model for the data.
        if provider := cls.network_manager.active_provider:
            ecosystem = provider.network.ecosystem
        else:
            logger.warning(
                "Given raw-transaction data when not connected to any provider. "
                "Network is unknown. Assuming EVM-like transaction model."
            )
            ecosystem = cls.network_manager.ethereum

        return ecosystem.create_transaction(**value)

    @cached_property
    def debug_logs_typed(self) -> list[tuple[Any]]:
        """Return any debug log data outputted by the transaction."""
        return []

    @cached_property
    def debug_logs_lines(self) -> list[str]:
        """
        Return any debug log data outputted by the transaction as strings suitable for printing
        """
        return [" ".join(map(str, ln)) for ln in self.debug_logs_typed]

    @property
    def error(self) -> TransactionError | None:
        return self._error

    @error.setter
    def error(self, value: TransactionError):
        self._error = value

    def show_debug_logs(self):
        """
        Output debug logs to logging system
        """
        for ln in self.debug_logs_lines:
            logger.info(f"[DEBUG-LOG] {ln}")

    @property
    def failed(self) -> bool:
        """
        Whether the receipt represents a failing transaction.
        Ecosystem plugins override this property when their receipts
        are able to be failing.
        """
        return False

    @property
    def confirmed(self) -> bool:
        """
        ``True`` when the number of confirmations is equal or greater
        to the required amount of confirmations.
        """
        return self._confirmations_occurred == self.required_confirmations

    @property
    @abstractmethod
    def total_fees_paid(self) -> int:
        """
        The total amount of fees paid for the transaction.
        """

    @property
    @abstractmethod
    def ran_out_of_gas(self) -> bool:
        """
        Check if a transaction ran out of gas and failed.

        Returns:
            bool:  ``True`` when the transaction failed and used the
            same amount of gas as the given ``gas_limit``.
        """

    @cached_property
    def trace(self) -> "TraceAPI":
        """
        The :class:`~ape.api.trace.TraceAPI` of the transaction.
        """
        return self.provider.get_transaction_trace(self.txn_hash)

    @property
    def _explorer(self) -> "ExplorerAPI | None":
        return self.provider.network.explorer

    @property
    def _block_time(self) -> int:
        return self.provider.network.block_time

    @property
    def _confirmations_occurred(self) -> int:
        latest_block = self.provider.get_block("latest")

        if latest_block.number is None:
            return 0

        return latest_block.number - self.block_number

    @cached_property
    def block(self) -> "BlockAPI":
        return self.chain_manager.blocks[self.block_number]

    @property
    def timestamp(self) -> int:
        return self.block.timestamp

    @property
    def datetime(self) -> "datetime_type":
        return self.block.datetime

    @cached_property
    def events(self) -> "ContractLogContainer":
        """
        All the events that were emitted from this call.
        """

        return self.decode_logs()  # Decodes all logs by default.

    @abstractmethod
    def decode_logs(
        self,
        abi: ("list[EventABI | ContractEvent] | EventABI | ContractEvent | None") = None,
    ) -> "ContractLogContainer":
        """
        Decode the logs on the receipt.

        Args:
            abi (``EventABI``): The ABI of the event to decode into logs.

        Returns:
            list[:class:`~ape.types.ContractLog`]
        """

    def raise_for_status(self) -> NoReturn | None:
        """
        Handle provider-specific errors regarding a non-successful
        :class:`~api.providers.TransactionStatusEnum`.
        """

    def await_confirmations(self) -> "ReceiptAPI":
        """
        Wait for a transaction to be considered confirmed.

        Returns:
            :class:`~ape.api.ReceiptAPI`: The receipt that is now confirmed.
        """
        # NOTE: Even when required_confirmations is `0`, we want to wait for the nonce to
        #   increment. Otherwise, users may end up with invalid nonce errors in tests.
        self._await_sender_nonce_increment()
        if self.required_confirmations == 0 or self._check_error_status() or self.confirmed:
            return self

        # Confirming now.
        self._log_submission()
        self._await_confirmations()
        return self

    def _await_sender_nonce_increment(self):
        if not self.sender:
            return

        iterations_timeout = 20
        iteration = 0
        sender_nonce = self.provider.get_nonce(self.sender)
        while sender_nonce == self.nonce:
            time.sleep(1)
            sender_nonce = self.provider.get_nonce(self.sender)
            iteration += 1
            if iteration != iterations_timeout:
                continue

            tx_err = TransactionError("Timeout waiting for sender's nonce to increase.")
            self.error = tx_err
            if self.transaction.raise_on_revert:
                raise tx_err
            else:
                break

    def _log_submission(self):
        if explorer_url := self._explorer and self._explorer.get_transaction_url(self.txn_hash):
            log_message = f"Submitted {explorer_url}"
        else:
            log_message = f"Submitted {self.txn_hash}"

        logger.info(log_message)

    def _check_error_status(self) -> bool:
        try:
            self.raise_for_status()
        except TransactionError:
            # Skip waiting for confirmations when the transaction has failed.
            return True

        return False

    def _await_confirmations(self):
        if self.required_confirmations <= 0:
            return

        with ConfirmationsProgressBar(self.required_confirmations) as progress_bar:
            while not self.confirmed:
                confirmations_occurred = self._confirmations_occurred
                if confirmations_occurred >= self.required_confirmations:
                    break

                progress_bar.confs = confirmations_occurred
                time_to_sleep = int(self._block_time / 2)
                time.sleep(time_to_sleep)

    @property
    def method_called(self) -> "MethodABI | None":
        """
        The method ABI of the method called to produce this receipt.
        """
        return None

    @cached_property
    def return_value(self) -> Any:
        """
        Obtain the final return value of the call. Requires tracing to function,
        since this is not available from the receipt object.
        """
        if trace := self.trace:
            ret_val = trace.return_value
            return ret_val[0] if isinstance(ret_val, tuple) and len(ret_val) == 1 else ret_val

        return None

    @property
    @raises_not_implemented
    def source_traceback(self) -> "SourceTraceback":  # type: ignore[empty-body]
        """
        A Pythonic style traceback for both failing and non-failing receipts.
        Requires a provider that implements
        :meth:~ape.api.providers.ProviderAPI.get_transaction_trace`.
        """

    @raises_not_implemented
    def show_trace(self, verbose: bool = False, file: IO[str] = sys.stdout):
        """
        Display the complete sequence of contracts and methods called during
        the transaction.

        Args:
            verbose (bool): Set to ``True`` to include more information.
            file (IO[str]): The file to send output to. Defaults to stdout.
        """

    @raises_not_implemented
    def show_gas_report(self, file: IO[str] = sys.stdout):
        """
        Display a gas report for the calls made in this transaction.
        """

    @raises_not_implemented
    def show_source_traceback(self):
        """
        Show a receipt traceback mapping to lines in the source code.
        Only works when the contract type and source code are both available,
        like in local projects.
        """

    @raises_not_implemented
    def show_events(self):
        """
        Show the events from the receipt.
        """

    def track_gas(self):
        """
        Track this receipt's gas in the on-going session gas-report.
        Requires using a provider that supports transaction traces
        to get full data. Else, is limited to receipt-level data.
        This gets called when running tests with the ``--gas`` flag.
        """
        address = self.receiver or self.contract_address
        if not address or not self._test_runner:
            return

        if self.provider.supports_tracing and (trace := self.trace):
            tracker = self._test_runner.gas_tracker
            tracker.append_gas(trace, address)

        elif (
            (contract_type := self.chain_manager.contracts.get(address))
            and contract_type.source_id
            and (method := self.method_called)
        ):
            # Can only track top-level gas.
            if contract := self.local_project._create_contract_source(contract_type):
                self._test_runner.gas_tracker.append_toplevel_gas(contract, method, self.gas_used)

    def track_coverage(self):
        """
        Track this receipt's source code coverage in the on-going
        session coverage report. Requires using a provider that supports
        transaction traces to track full coverage. Else, is limited
        to receipt-level tracking. This gets called when running tests with
        the ``--coverage`` flag.
        """

        if not self.network_manager.active_provider or not self._test_runner:
            return

        if not (address := self.receiver):
            # NOTE: Deploy txns are currently not tracked!
            return

        tracker = self._test_runner.coverage_tracker
        if self.provider.supports_tracing and (traceback := self.source_traceback):
            if len(traceback) > 0:
                tracker.cover(traceback)

        elif method := self.method_called:
            # Unable to track detailed coverage like statement or branch
            # The user will receive a warning at the end regarding this.
            # At the very least, we can track function coverage.
            contract_type = self.chain_manager.contracts.get(address)
            if not contract_type or not contract_type.source_id:
                return

            if contract := self.local_project._create_contract_source(contract_type):
                tracker.hit_function(contract, method)



import json
import os
import shutil
import subprocess
import sys
import time
from collections.abc import Callable, Sequence
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Optional, Union

import pytest
from click.testing import CliRunner

import ape
from ape.logging import LogLevel, logger
from ape.managers.project import LocalProject, Project
from ape.pytest.config import ConfigWrapper
from ape.pytest.gas import GasTracker
from ape.types.address import AddressType
from ape.types.units import CurrencyValue
from ape.utils.basemodel import only_raise_attribute_error
from ape.utils.misc import ZERO_ADDRESS
from ape.utils.os import create_tempdir
from ape.utils.testing import DEFAULT_TEST_CHAIN_ID

# Needed to test tracing support in core `ape test` command.
pytest_plugins = ["pytester"]
GETH_URI = "http://127.0.0.1:5550"
ALIAS = "__FUNCTIONAL_TESTS_ALIAS__"
geth_process_test = pytest.mark.xdist_group(name="geth-tests")
explorer_test = pytest.mark.xdist_group(name="explorer-tests")

# Ensure we don't persist any .ape data or using existing.

_DATA_FOLDER_CTX = ape.config.isolate_data_folder(keep="packages")
DATA_FOLDER = _DATA_FOLDER_CTX.__enter__()
TESTS_FOLDER = Path(__file__).parent
FUNCTIONAL_TESTS_FOLDER = TESTS_FOLDER / "functional"
SHARED_CONTRACTS_FOLDER = FUNCTIONAL_TESTS_FOLDER / "data" / "contracts"


EXPECTED_MYSTRUCT_C = 244
"""Hardcoded everywhere in contracts."""


def pytest_addoption(parser):
    parser.addoption(
        "--includepip", action="store_true", help="run tests that depend on pip install operations"
    )


def pytest_runtest_setup(item):
    if "pip" in item.keywords and not item.config.getoption("--includepip"):
        pytest.skip("need --includepip option to run this test")


@pytest.fixture(autouse=True)
def setenviron(monkeypatch):
    """
    Sets the APE_TESTING environment variable during tests.

    With this variable set fault handling and IPython command history logging
    will be disabled in the ape console.
    """
    monkeypatch.setenv("APE_TESTING", "1")


@pytest.fixture(scope="session", autouse=True)
def clean_temp_data_folder():
    yield
    _DATA_FOLDER_CTX.__exit__(None, None, None)


@pytest.fixture(scope="session", autouse=True)
def start_dir():
    return os.getcwd()


@pytest.fixture(autouse=True)
def validate_cwd(start_dir):
    # Handle weird issues with cwd breaking everything.
    # Possibly dur to chdir to a tempdir and then it gets deleted. my guess.
    # TODO: Find root cause and fix there.
    try:
        os.getcwd()
    except Exception:
        # Change back to project root, hopefully.
        os.chdir(start_dir)


@pytest.fixture(scope="session")
def config():
    return ape.config


@pytest.fixture(scope="session")
def conversion_manager(chain):
    return chain.conversion_manager


@pytest.fixture(scope="session")
def convert(conversion_manager):
    return conversion_manager.convert


@pytest.fixture(scope="session")
def data_folder(config):
    return DATA_FOLDER


@pytest.fixture(scope="session")
def projects_path():
    return Path(__file__).parent / "integration" / "cli" / "projects"


@pytest.fixture(scope="session")
def with_dependencies_project_path(projects_path):
    return projects_path / "with-dependencies"


@pytest.fixture(scope="session")
def plugin_manager():
    return ape.networks.plugin_manager


@pytest.fixture(scope="session")
def account_manager():
    # NOTE: `accounts` fixture comes with ape_test as the test-accounts.
    return ape.accounts


@pytest.fixture(scope="session")
def compilers():
    return ape.compilers


@pytest.fixture(scope="session")
def owner(accounts):
    return accounts[0]


@pytest.fixture(scope="session")
def sender(accounts):
    return accounts[1]


@pytest.fixture(scope="session")
def receiver(accounts):
    return accounts[2]


@pytest.fixture(scope="session")
def not_owner(accounts):
    return accounts[3]


@pytest.fixture(scope="session")
def helper(accounts):
    return accounts[4]


@pytest.fixture(scope="session")
def mystruct_c():
    return CurrencyValue(EXPECTED_MYSTRUCT_C)


@pytest.fixture
def signer(accounts):
    return accounts[5]


@pytest.fixture
def geth_account(accounts):
    return accounts[6]


@pytest.fixture
def geth_second_account(accounts):
    return accounts[7]


@pytest.fixture(scope="session")
def keyparams():
    # NOTE: password is 'asdf1234'
    return {
        "address": "f39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        "crypto": {
            "cipher": "aes-128-ctr",
            "cipherparams": {"iv": "229df0a8949798c192caf21531b64a01"},
            "ciphertext": "c68e03a33ab139a4822f578d76452658c13a0ea370f3c997651613dea8925483",
            "kdf": "scrypt",
            "kdfparams": {
                "dklen": 32,
                "n": 262144,
                "r": 8,
                "p": 1,
                "salt": "e1f01ece5afa9819e0ff6c1761737c68",
            },
            "mac": "10ee5db98f1a653c9bda7657f3b3b8bd55dd2fec93936e6b1783af912f9167c2",
        },
        "id": "1af390c5-c4cf-46d0-9341-5374e1a84959",
        "version": 3,
    }


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def networks_disconnected():
    provider = ape.networks.active_provider
    ape.networks.active_provider = None

    try:
        yield ape.networks
    finally:
        ape.networks.active_provider = provider


@pytest.fixture
def ethereum(networks):
    return networks.ethereum


@pytest.fixture(autouse=True)
def eth_tester_provider(ethereum):
    # NOTE: Ensure it uses the default instance of eth-tester.
    with ethereum.local.use_provider(
        "test", provider_settings={"chain_id": DEFAULT_TEST_CHAIN_ID}
    ) as provider:
        yield provider


@pytest.fixture
def mock_provider(mock_web3, eth_tester_provider):
    web3 = eth_tester_provider.web3
    eth_tester_provider._web3 = mock_web3
    yield eth_tester_provider
    eth_tester_provider._web3 = web3


@pytest.fixture
def networks_connected_to_tester(eth_tester_provider):
    return eth_tester_provider.network_manager


@pytest.fixture
def geth_provider(networks):
    if (
        not networks.active_provider
        or networks.provider.name != "node"
        or networks.network.name != "local"
        or not networks.provider.is_connected
        or getattr(networks.provider, "uri", "") != GETH_URI
    ):
        test_acct_100 = "0x63c7f11162dBFC374DC6f5C0B3Aa26C618846a85"
        with networks.ethereum.local.use_provider(
            "node", provider_settings={"uri": GETH_URI, "extra_funded_accounts": [test_acct_100]}
        ) as provider:
            yield provider
    else:
        yield networks.provider


@pytest.fixture
def empty_data_folder():
    # Avoid user's global ape-config data.
    if "global_config" in (ape.config.__dict__ or {}):
        del ape.config.__dict__["global_config"]

    packages = DATA_FOLDER / "packages"
    with create_tempdir() as temp_dir:
        temp_packages = temp_dir / "packages"
        if packages.is_dir():
            shutil.copytree(packages, temp_packages)

        shutil.rmtree(DATA_FOLDER, ignore_errors=True)
        DATA_FOLDER.mkdir(parents=True, exist_ok=True)
        if temp_packages.is_dir():
            shutil.copytree(temp_packages, DATA_FOLDER / "packages")

    shutil.rmtree(temp_packages, ignore_errors=True)


@pytest.fixture
def keyfile_account(owner, keyparams, temp_keyfile_account_ctx):
    with temp_keyfile_account_ctx(ALIAS, keyparams, owner) as account:
        # Ensure starts off locked.
        account.lock()
        yield account


@pytest.fixture
def temp_keyfile_account_ctx():
    @contextmanager
    def _temp_keyfile_account(alias: str, keyparams, sender):
        accts_folder = DATA_FOLDER / "accounts"
        accts_folder.mkdir(parents=True, exist_ok=True)
        test_keyfile_path = accts_folder / f"{alias}.json"

        if test_keyfile_path.is_file():
            account = ape.accounts.load(ALIAS)
        else:
            account = _make_keyfile_account(accts_folder, alias, keyparams, sender)

        try:
            yield account
        finally:
            if test_keyfile_path.is_file():
                test_keyfile_path.unlink()

    return _temp_keyfile_account


def _make_keyfile_account(base_path: Path, alias: str, params: dict, funder):
    test_keyfile_path = base_path / f"{alias}.json"

    if test_keyfile_path.is_file():
        # Corrupted from a previous test
        test_keyfile_path.unlink()

    test_keyfile_path.write_text(json.dumps(params), encoding="utf8")
    acct = ape.accounts.load(alias)
    funder.transfer(acct, "25 ETH")  # Auto-fund this account
    return acct


def skip_if_plugin_installed(*plugin_names: str):
    """
    A simple decorator for skipping a test if a plugin is installed.
    **NOTE**: For performance reasons, this method is not very good.
    It only works for common ApeWorX supported plugins and is only
    meant for assisting testing in Core (NOT a public utility).
    """
    names = [n.lower().replace("-", "_").replace("ape_", "") for n in plugin_names]
    msg_f = "Cannot run this test when plugin '{}' installed."

    def wrapper(fn):
        for name in names:
            # Compilers
            if name in ("solidity", "vyper"):
                compiler = ape.compilers.get_compiler(name)
                if compiler:

                    def test_skip_from_compiler(*args, name=name, **kwargs):
                        pytest.mark.skip(msg_f.format(name))

                    # NOTE: By returning a function, we avoid a collection warning.
                    return test_skip_from_compiler

            # Converters
            elif name in ("ens",):
                address_converters = [
                    type(n).__name__ for n in ape.chain.conversion_manager._converters[AddressType]
                ]
                if any(x.startswith(name.upper()) for x in address_converters):

                    def test_skip_from_converter(name=name):
                        pytest.mark.skip(msg_f.format(name))

                    return test_skip_from_converter
        # noop
        return fn

    return wrapper


@pytest.fixture
def zero_address():
    return ZERO_ADDRESS


@pytest.fixture
def ape_caplog(caplog):
    class ApeCaplog:
        def __init__(self, caplog_level: LogLevel = LogLevel.WARNING):
            self.level = caplog_level
            self.messages_at_start = list(caplog.messages)
            self.set_levels(caplog_level)

        @only_raise_attribute_error
        def __getattr__(self, name: str) -> Any:
            return getattr(caplog, name)

        @contextmanager
        def at_level(self, level: LogLevel):
            original = self.level
            self.set_levels(level)
            yield
            self.set_levels(original)

        @property
        def fail_message(self) -> str:
            if caplog.messages:
                last_message = caplog.messages[-1]
                return f"Actual last message: {last_message}"

            elif self.messages_at_start:
                return (
                    f"Failed to detect logs. "
                    f"However, we did have logs before the operation: "
                    f"{', '.join(self.messages_at_start)}"
                )

            else:
                return "No logs found!"

        @property
        def head(self) -> str:
            """
            A str representing the latest logged line.
            Initialized to empty str.
            """
            return caplog.messages[-1] if len(caplog.messages) else ""

        def set_levels(self, level: LogLevel = LogLevel.WARNING):
            self.level = level
            logger.set_level(level)
            caplog.set_level(level)

        def assert_last_log(self, message: str):
            assert message in self.head, self.fail_message

        def assert_last_log_with_retries(
            self, op: Callable, message: str, tries: int = 2, delay: float = 5.0
        ):
            times_tried = 0
            return_value = None
            while times_tried <= tries:
                result = op()

                # Only save the first return value.
                if return_value is None and result is not None:
                    return_value = result

                times_tried += 1
                if message in self.head:
                    return return_value

                time.sleep(delay)

                # Reset levels in case they got switched.
                self.set_levels()
                logger.set_level(LogLevel.ERROR)
                caplog.set_level(LogLevel.WARNING)

            pytest.fail(self.fail_message)

    return ApeCaplog()


@pytest.fixture
def mock_home_directory(tmp_path):
    original_home = Path.home()
    Path.home = lambda: tmp_path  # type: ignore[method-assign]
    yield tmp_path
    Path.home = lambda: original_home  # type: ignore[method-assign]


class SubprocessRunner:
    """
    Same CLI commands are better tested using a python subprocess,
    such as `ape test` commands because duplicate pytest main methods
    do not run well together, or `ape plugins` commands, which may
    modify installed plugins.
    """

    def __init__(self, root_cmd: Sequence[str] | None = None, data_folder: Path | None = None):
        self.root_cmd = root_cmd or []
        self.data_folder = data_folder

    def invoke(
        self,
        *subcommand: str,
        input=None,
        timeout: int = 40,
        env: dict | None = None,
    ):
        subcommand = subcommand or ()
        cmd_ls = [*self.root_cmd, *subcommand]

        env = {**dict(os.environ), **(env or {})}
        if self.data_folder:
            env["APE_DATA_FOLDER"] = str(self.data_folder)

        completed_process = subprocess.run(
            cmd_ls,
            capture_output=True,
            env=env,
            input=input,
            text=True,
            timeout=timeout,
            check=False,
        )
        result = SubprocessResult(completed_process)
        sys.stdin = sys.__stdin__
        return result


class ApeSubprocessRunner(SubprocessRunner):
    """
    Subprocess runner for Ape-specific commands.
    """

    def __init__(
        self,
        root_cmd: str | Sequence[str] | None = None,
        data_folder: Path | None = None,
    ):
        ape_path = Path(sys.executable).parent / "ape"

        root = root_cmd or ()
        if isinstance(root, str):
            root = (root,)

        super().__init__([str(ape_path), *root], data_folder=data_folder)
        self.project = None

    def invoke(self, *subcommand: str, input=None, timeout: int = 40, env: dict | None = None):
        if self.project:
            try:
                here = os.getcwd()
            except Exception:
                here = None

            os.chdir(f"{self.project.path}")

        else:
            here = None

        result = super().invoke(*subcommand, input=input, timeout=timeout, env=env)
        if here:
            os.chdir(here)

        return result


class SubprocessResult:
    def __init__(self, completed_process: subprocess.CompletedProcess):
        self._completed_process = completed_process

    @property
    def exit_code(self) -> int:
        return self._completed_process.returncode

    @property
    def output(self) -> str:
        return self._completed_process.stdout or self._completed_process.stderr


CUSTOM_NETWORK_0 = "apenet"
CUSTOM_NETWORK_CHAIN_ID_0 = 944898498948934528628
CUSTOM_NETWORK_1 = "apenet1"
CUSTOM_NETWORK_CHAIN_ID_1 = 944898498948934528629
CUSTOM_BLOCK_TIME = 123


def _make_net(name: str, chain_id: int, **kwargs) -> dict:
    return {"name": name, "chain_id": chain_id, "ecosystem": "ethereum", **kwargs}


CUSTOM_NETWORKS_CONFIG = {
    "networks": {
        "custom": [
            _make_net(CUSTOM_NETWORK_0, CUSTOM_NETWORK_CHAIN_ID_0),
            _make_net(CUSTOM_NETWORK_1, CUSTOM_NETWORK_CHAIN_ID_1),
        ]
    }
}


@pytest.fixture(scope="session")
def custom_networks_config_dict():
    return CUSTOM_NETWORKS_CONFIG


@pytest.fixture(scope="session")
def custom_network_name_0():
    return CUSTOM_NETWORK_0


@pytest.fixture(scope="session")
def custom_network_name_1():
    return CUSTOM_NETWORK_1


@pytest.fixture(scope="session")
def custom_network_chain_id_0():
    return CUSTOM_NETWORK_CHAIN_ID_0


@pytest.fixture(scope="session")
def custom_network_chain_id_1():
    return CUSTOM_NETWORK_CHAIN_ID_1


@pytest.fixture
def custom_network(ethereum, project, custom_networks_config_dict):
    with project.temp_config(**custom_networks_config_dict):
        yield ethereum.apenet


@pytest.fixture
def config_wrapper(mocker):
    return ConfigWrapper(mocker.MagicMock())


@pytest.fixture
def gas_tracker(config_wrapper):
    return GasTracker(config_wrapper)


@pytest.fixture(scope="session")
def shared_contracts_folder():
    return SHARED_CONTRACTS_FOLDER


@pytest.fixture
def smaller_project(with_dependencies_project_path):
    """
    Useful to avoid compiling the massive main test contracts.
    """
    return Project(with_dependencies_project_path)


@pytest.fixture
def small_temp_project(smaller_project):
    with smaller_project.isolate_in_tempdir() as temp_project:
        yield temp_project


@pytest.fixture
def temp_project(project):
    with project.isolate_in_tempdir() as temp_project:
        yield temp_project


@pytest.fixture
def empty_project():
    with create_tempdir() as empty:
        yield LocalProject(empty)


@pytest.fixture
def geth_contract(geth_account, project, geth_provider):
    return geth_account.deploy(project.SolidityContract, 0)



import asyncio
import functools
import inspect
import json
import sys

if sys.version_info.minor >= 11:
    # 3.11 or greater
    # NOTE: type-ignore is for when running mypy on python versions < 3.11
    import tomllib  # type: ignore[import-not-found]
else:
    import toml as tomllib  # type: ignore[no-redef]

from asyncio import gather
from collections.abc import Coroutine, Mapping
from datetime import datetime, timezone
from functools import cached_property, singledispatchmethod, wraps
from importlib.metadata import PackageNotFoundError, distributions
from importlib.metadata import version as version_metadata
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar, cast

import yaml
from eth_keys import keys  # type: ignore
from eth_pydantic_types import HexBytes
from eth_utils import is_0x_prefixed
from packaging.specifiers import SpecifierSet

from ape.exceptions import APINotImplementedError
from ape.logging import logger
from ape.utils.os import expand_environment_variables

if TYPE_CHECKING:
    from ape.types.address import AddressType


EMPTY_BYTES32 = HexBytes("0x0000000000000000000000000000000000000000000000000000000000000000")
ZERO_ADDRESS: "AddressType" = cast("AddressType", "0x0000000000000000000000000000000000000000")
DEFAULT_TRANSACTION_ACCEPTANCE_TIMEOUT = 120
DEFAULT_LOCAL_TRANSACTION_ACCEPTANCE_TIMEOUT = 20
DEFAULT_LIVE_NETWORK_BASE_FEE_MULTIPLIER = 1.4
DEFAULT_TRANSACTION_TYPE = 0
DEFAULT_MAX_RETRIES_TX = 20
LOCAL_NETWORK_NAME = "local"
SOURCE_EXCLUDE_PATTERNS = (
    ".build",
    ".cache",
    ".DS_Store",
    ".git",
    ".gitkeep",
    "*.adoc",
    "*.css",
    "*.html",
    "*.md",
    "*.pdf",
    "*.py*",
    "*.rst",
    "*.txt",
    "*package.json",
    "*package-lock.json",
    "*tsconfig.json",
    "ape-config.yaml",
    "py.typed",
)


_python_version: str = (
    f"{sys.version_info.major}.{sys.version_info.minor}"
    f".{sys.version_info.micro} {sys.version_info.releaselevel}"
)


@functools.cache
def _get_distributions(pkg_name: str | None = None) -> list:
    """
    Get a mapping of top-level packages to their distributions.
    """

    distros = []
    all_distros = distributions()
    for dist in all_distros:
        package_names = (dist.read_text("top_level.txt") or "").split()
        for name in package_names:
            if pkg_name is None or name == pkg_name:
                distros.append(dist)

    return distros


def pragma_str_to_specifier_set(pragma_str: str) -> SpecifierSet | None:
    """
    Convert the given pragma str to a ``packaging.version.SpecifierSet``
    if possible.

    Args:
        pragma_str (str): The str to convert.

    Returns:
        SpecifierSet | None
    """

    pragma_parts = iter([x.strip(" ,") for x in pragma_str.split(" ")])

    def _to_spec(item: str) -> str:
        item = item.replace("^", "~=")
        if item and item[0].isnumeric():
            return f"=={item}"
        elif item and len(item) >= 2 and item[0] == "=" and item[1] != "=":
            return f"={item}"

        return item

    pragma_parts_fixed = []
    builder = ""
    for sub_part in pragma_parts:
        parts_to_handle: list[str] = []
        if "," in sub_part:
            sub_sub_parts = [x.strip() for x in sub_part.split(",")]
            if len(sub_sub_parts) > 2:
                # Very rare case.
                raise ValueError(f"Cannot handle pragma '{pragma_str}'.")

            if next_part := next(pragma_parts, None):
                parts_to_handle.extend((sub_sub_parts[0], f"{sub_sub_parts[-1]}{next_part}"))
            else:
                # Very rare case.
                raise ValueError(f"Cannot handle pragma '{pragma_str}'.")
        else:
            parts_to_handle.append(sub_part)

        for part in parts_to_handle:
            if not any(c.isnumeric() for c in part):
                # Handle pragma with spaces between constraint and values
                # like `>= 0.6.0`.
                builder += part
                continue
            elif builder:
                spec = _to_spec(f"{builder}{part}")
                builder = ""
            else:
                spec = _to_spec(part)

            pragma_parts_fixed.append(spec)

    try:
        return SpecifierSet(",".join(pragma_parts_fixed))
    except ValueError as err:
        logger.error(str(err))
        return None


def get_package_version(obj: Any) -> str:
    """
    Get the version of a single package.

    Args:
        obj: object to search inside for ``__version__``.

    Returns:
        str: version string.
    """
    # If value is already cached/static
    if hasattr(obj, "__version__"):
        return obj.__version__

    # NOTE: In case where don't pass a module name
    if not isinstance(obj, str) and hasattr(obj, "__name__"):
        obj = obj.__name__

    elif not isinstance(obj, str):
        try:
            str_value = f"{obj}"
        except Exception:
            str_value = "<obj>"

        logger.warning(f"Type issue: Unknown if properly handled {str_value}")

        # Treat as no version found.
        return ""

    # Reduce module string to base package
    # NOTE: Assumed that string input is module name e.g. `__name__`
    pkg_name = obj.partition(".")[0]

    # NOTE: In case the distribution and package name differ
    dists = _get_distributions(pkg_name)
    if dists:
        pkg_name = dists[0].metadata["Name"]
    try:
        return str(version_metadata(pkg_name))

    except PackageNotFoundError:
        # NOTE: Must handle empty string result here
        return ""


__version__: str = get_package_version(__name__)


def load_config(path: Path, expand_envars=True, must_exist=False) -> dict:
    """
    Load a configuration file into memory.
    A file at the given path must exist or else it will throw ``OSError``.
    The configuration file must be a `.json` or `.yaml` or else it will throw ``TypeError``.

    Args:
        path (str): path to filesystem to find.
        expand_envars (bool): ``True`` the variables in path
                                are able to expand to show full path.
        must_exist (bool): ``True`` will be set if the configuration file exist
                                and is able to be load.

    Returns:
        dict: Configured settings parsed from a config file.
    """
    if path.is_file():
        contents = path.read_text()
        if expand_envars:
            contents = expand_environment_variables(contents)

        if path.name == "pyproject.toml":
            pyproject_toml = tomllib.loads(contents)
            config = pyproject_toml.get("tool", {}).get("ape", {})

            # Utilize [project] for some settings.
            if project_settings := pyproject_toml.get("project"):
                if "name" not in config and "name" in project_settings:
                    config["name"] = project_settings["name"]

        elif path.suffix in (".json",):
            config = json.loads(contents)
        elif path.suffix in (".yml", ".yaml"):
            config = yaml.safe_load(contents)
        else:
            raise TypeError(f"Cannot parse '{path.suffix}' files!")

        return config or {}

    elif must_exist:
        raise OSError(f"{path} does not exist!")

    else:
        return {}


def gas_estimation_error_message(tx_error: Exception) -> str:
    """
    Get an error message containing the given error and an explanation of how the
    gas estimation failed, as in :class:`ape.api.providers.ProviderAPI` implementations.

    Args:
        tx_error (Exception): The error that occurred when trying to estimate gas.

    Returns:
        str: An error message explaining that the gas failed and that the transaction
        will likely revert.
    """
    txn_error_text = str(tx_error)
    if txn_error_text.endswith("."):
        # Strip period from initial error so it looks better.
        txn_error_text = txn_error_text[:-1]

    return (
        f"Gas estimation failed: '{txn_error_text}'. This transaction will likely revert. "
        "If you wish to broadcast, you must set the gas limit manually."
    )


def extract_nested_value(root: Mapping, *args: str) -> dict | None:
    """
    Dig through a nested ``dict`` using the given keys and return the
    last-found object.

    Usage example::

            >>> extract_nested_value({"foo": {"bar": {"test": "VALUE"}}}, "foo", "bar", "test")
            'VALUE'

    Args:
        root (dict): Nested keys to form arguments.

    Returns:
        dict, optional: The final value if it exists
        else ``None`` if the tree ends at any point.
    """
    current_value: Any = root
    for arg in args:
        if not hasattr(current_value, "get"):
            return None

        current_value = current_value.get(arg)

    return current_value


def add_padding_to_strings(
    str_list: list[str],
    extra_spaces: int = 0,
    space_character: str = " ",
) -> list[str]:
    """
    Append spacing to each string in a list of strings such that
    they all have the same length.

    Args:
        str_list (list[str]): The list of strings that need padding.
        extra_spaces (int): Optionally append extra spacing. Defaults to ``0``.
        space_character (str): The character to use in the padding. Defaults to ``" "``.

    Returns:
        list[str]: A list of equal-length strings with padded spaces.
    """

    if not str_list:
        return []

    longest_item = len(max(str_list, key=len))
    spaced_items = []

    for value in str_list:
        spacing = (longest_item - len(value) + extra_spaces) * space_character
        spaced_items.append(f"{value}{spacing}")

    return spaced_items


def raises_not_implemented(fn):
    """
    Decorator for raising helpful not implemented error.
    """

    def inner(*args, **kwargs):
        raise _create_raises_not_implemented_error(fn)

    # This is necessary for doc strings to show up with sphinx
    inner.__doc__ = fn.__doc__
    return inner


def _create_raises_not_implemented_error(fn):
    return APINotImplementedError(
        f"Attempted to call method '{fn.__qualname__}', method not supported."
    )


def to_int(value: Any) -> int:
    """
    Convert the given value, such as hex-strs or hex-bytes, to an integer.
    """

    if isinstance(value, int):
        return value
    elif isinstance(value, str):
        return int(value, 16) if is_0x_prefixed(value) else int(value)
    elif isinstance(value, bytes):
        return int.from_bytes(value, "big")

    raise ValueError(f"cannot convert {value!r} to int")


def run_until_complete(*item: Any) -> Any:
    """
    Completes the given coroutine and returns its value.

    Args:
        *item (Any): A coroutine or any return value from an async method. If
          not given a coroutine, returns the given item. Provide multiple
          coroutines to run tasks in parallel.

    Returns:
        (Any): The value that results in awaiting the coroutine.
        Else, ``item`` if ``item`` is not a coroutine. If given multiple coroutines,
        returns the result from ``asyncio.gather``.
    """

    items = list(item)
    if not items:
        return None

    elif not isinstance(items[0], Coroutine):
        # Method was marked `async` but didn't return a coroutine.
        # This happens in some `web3.py` methods.
        return items if len(items) > 1 else items[0]

    # Run all coroutines async.
    task = gather(*items, return_exceptions=True) if len(items) > 1 else items[0]
    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(task)
    return result


def nonreentrant(key_fn):
    def inner(f):
        locks = set()

        @wraps(f)
        def wrapper(*args, **kwargs):
            key = key_fn(*args, **kwargs)
            if key in locks:
                raise RecursionError(f"nonreentrant {f.__qualname__}:{key}")
            locks.add(key)
            try:
                return f(*args, **kwargs)
            finally:
                locks.discard(key)

        return wrapper

    return inner


def get_current_timestamp_ms() -> int:
    """
    Get the current UNIX timestamp in milliseconds.

    Returns:
        int
    """
    return round(datetime.now(tz=timezone.utc).timestamp() * 1000)


def is_evm_precompile(address: str) -> bool:
    """
    Returns ``True`` if the given address string is a known
    Ethereum pre-compile address.

    Args:
        address (str):

    Returns:
        bool
    """
    try:
        address = address.replace("0x", "")
        return 0 < sum(int(x) for x in address) < 10
    except Exception:
        return False


def is_zero_hex(address: str) -> bool:
    """
    Returns ``True`` if the hex str is only zero.
    **NOTE**: Empty hexes like ``"0x"`` are considered zero.

    Args:
        address (str): The address to check.

    Returns:
        bool
    """

    try:
        if addr := address.replace("0x", ""):
            return sum(int(x) for x in addr) == 0
        else:
            # "0x" counts as zero.
            return True

    except Exception:
        return False


def _dict_overlay(mapping: dict[str, Any], overlay: dict[str, Any], depth: int = 0):
    """Overlay given overlay structure on a dict"""
    for key, value in overlay.items():
        if isinstance(value, dict):
            if key not in mapping:
                mapping[key] = dict()
            _dict_overlay(mapping[key], value, depth + 1)
        else:
            mapping[key] = value
    return mapping


def log_instead_of_fail(default: Any | None = None):
    """
    A decorator for logging errors instead of raising.
    This is useful for methods like __repr__ which shouldn't fail.
    """

    def wrapper(fn):
        @functools.wraps(fn)
        def wrapped(*args, **kwargs):
            try:
                if args and isinstance(args[0], type):
                    return fn(*args, **kwargs)
                else:
                    return fn(*args, **kwargs)

            except Exception as err:
                logger.error(str(err))
                if default:
                    return default

        return wrapped

    return wrapper


_MOD_T = TypeVar("_MOD_T")


def as_our_module(cls_or_def: _MOD_T, doc_str: str | None = None) -> _MOD_T | None:
    """
    Ape sometimes reclaims definitions from other packages, such as
    class:`~ape.types.signatures.SignableMessage`). When doing so, the doc str
    may be different than ours, and the module may still refer to
    the original package. This method steals the given class as-if
    it were ours. Logic borrowed from starknet-py.
    https://github.com/software-mansion/starknet.py/blob/0.10.1-alpha/starknet_py/utils/docs.py#L10-L24

    Args:
        cls_or_def (_MOD_T): The class or definition to borrow.
        doc_str (str): Optionally change the doc string.

    Returns:
        The borrowed-version of the class or definition.
    """
    if cls_or_def is None:
        return cls_or_def

    elif doc_str is not None:
        cls_or_def.__doc__ = doc_str

    frame = inspect.stack()[1]
    if module := inspect.getmodule(frame[0]):
        cls_or_def.__module__ = module.__name__

    return cls_or_def


def derive_public_key(private_key: bytes) -> HexBytes:
    """
    Derive the public key for the given private key.

    Args:
        private_key (bytes): The private key.

    Returns:
        HexBytes: The public key.
    """
    pk = keys.PrivateKey(private_key)
    public_key = f"{pk.public_key}"[2:]
    return HexBytes(bytes.fromhex(public_key))


__all__ = [
    "LOCAL_NETWORK_NAME",
    "_dict_overlay",
    "cached_property",
    "extract_nested_value",
    "gas_estimation_error_message",
    "get_current_timestamp_ms",
    "get_package_version",
    "is_evm_precompile",
    "is_zero_hex",
    "load_config",
    "log_instead_of_fail",
    "nonreentrant",
    "pragma_str_to_specifier_set",
    "raises_not_implemented",
    "run_until_complete",
    "singledispatchmethod",
    "to_int",
]



from types import UnionType
from typing import Union, get_origin

import pytest

from ape.managers.converters import (
    AddressAPIConverter,
    BytesAddressConverter,
    HexConverter,
    HexIntConverter,
    IntAddressConverter,
    StringDecimalConverter,
    StringIntConverter,
)
from ape.types.address import AddressType
from ape.utils.basemodel import ManagerAccessMixin
from ape_ethereum._converters import WeiConversions

NAME_TO_CONVERTER = {
    "AddressAPI": AddressAPIConverter,
    "BytesAddress": BytesAddressConverter,
    "hex": HexConverter,
    "HexInt": HexIntConverter,
    "IntAddress": IntAddressConverter,
    "StringDecimal": StringDecimalConverter,
    "StringInt": StringIntConverter,
    "wei": WeiConversions,
}


@pytest.fixture
def conversion_manager():
    return ManagerAccessMixin.conversion_manager


@pytest.mark.parametrize("name", NAME_TO_CONVERTER)
def test_get_converter(name, conversion_manager):
    actual = conversion_manager.get_converter(name)
    expected = NAME_TO_CONVERTER[name]
    assert isinstance(actual, expected)


def test_get_converters_by_type(conversion_manager):
    converters = conversion_manager.get_converters_by_type(AddressType)
    for expected in (AddressAPIConverter, IntAddressConverter, BytesAddressConverter):
        assert any(isinstance(c, expected) for c in converters)


@pytest.mark.parametrize(
    "kwargs",
    [
        {"gas_limit": "0x2a"},
        {"receiver": "0x0000000000000000000000000000000000000001"},
    ],
)
def test_convert_method_kwargs_unwraps_pep604_optional(monkeypatch, conversion_manager, kwargs):
    """
    Ensure `ConversionManager.convert_method_kwargs()` unwraps `X | None` annotations
    before attempting conversion (Python 3.10+ union syntax).
    """

    seen: dict[str, object] = {}

    def fake_convert(_value, to_type):
        seen["to_type"] = to_type
        return 123

    monkeypatch.setattr(conversion_manager, "convert", fake_convert)
    actual = conversion_manager.convert_method_kwargs(kwargs)

    # We attempted conversion and used the converted value.
    key = next(iter(kwargs))
    assert actual[key] == 123

    # And the chosen type is not a union (i.e. None was stripped).
    assert "to_type" in seen
    origin = get_origin(seen["to_type"])
    assert origin not in (Union, UnionType)



def test_hash_same_as_python_hash(accounts):
    for account in accounts:
        # NOTE: According to CPython spec, `hash` is truncated according to OS bit width
        #       https://docs.python.org/3/library/functions.html#hash
        assert account.__hash__() == hash(account)


def test_address_set_membership(accounts):
    address_set = set(accounts)
    assert len(address_set) == len(accounts)
    assert accounts[0] in address_set

    new_set: set = {accounts[0], accounts[0]}
    assert len(new_set) == 1
    assert accounts[1] not in new_set


def test_address_in_dict(accounts):
    address_dict = {k: i for i, k in enumerate(accounts)}
    assert len(address_dict) == len(accounts)
    assert accounts[0] in address_dict
    assert address_dict[accounts[0]] == 0
